# coding=utf-8
from __future__ import print_function
import unittest
import ee
from .. import tools, indices
ee.Initialize()
from __init__ import TEST_CLOUD_IMAGES

p = ee.Geometry.Point([-71.69, -42.07])
image = ee.Image(TEST_CLOUD_IMAGES['S2'])
col = ee.ImageCollection('COPERNICUS/S2').filterBounds(p)

class TestAll(unittest.TestCase):
    def test(self):
        ndvi_i = indices.ndvi('B8', 'B4', addBand=False)(image)
        evi_i = indices.evi('B8', 'B4', 'B2', addBand=False)(image)
        nbr_i = indices.nbr('B8', 'B12', addBand=False)(image)
        nbr2_i = indices.nbr2('B11', 'B12', addBand=False)(image)

        values = tools.get_value(image, p, 10, side='client')

        NIR = float(values['B8'])
        SWIR1 = float(values['B11'])
        SWIR2 = float(values['B12'])
        RED = float(values['B4'])
        BLUE = float(values['B2'])
        GREEN = float(values['B3'])

        ndvi = (NIR-RED)/(NIR+RED)
        evi = (2.5)*((NIR-RED)/(NIR+(6*RED)-((7.5)*BLUE)+1))
        nbr = (NIR-SWIR2)/(NIR+SWIR2)
        nbr2 = (SWIR1-SWIR2)/(SWIR1+SWIR2)

        ndvi_from_i = tools.get_value(ndvi_i, p, 10, 'client')['NDVI']
        evi_from_i = tools.get_value(evi_i, p, 10, 'client')['EVI']
        nbr_from_i = tools.get_value(nbr_i, p, 10, 'client')['NBR']
        nbr2_from_i = tools.get_value(nbr2_i, p, 10, 'client')['NBR2']

        self.assertEqual(ndvi, ndvi_from_i)
        self.assertEqual(evi, evi_from_i)
        self.assertEqual(nbr, nbr_from_i)
        self.assertEqual(nbr2, nbr2_from_i)

class TestMap(unittest.TestCase):
    def test_ndvi(self):
        ndvi_col = col.map(indices.ndvi('B8','B4'))
        ndvi_i = ee.Image(ndvi_col.first())
        values = tools.get_value(ndvi_i, p, 10, side='client')
        NIR = float(values['B8'])
        RED = float(values['B4'])
        ndvi = (NIR-RED)/(NIR+RED)
        ndvi_from_i = tools.get_value(ndvi_i, p, 10, 'client')['NDVI']
        self.assertEqual(ndvi, ndvi_from_i)


# coding=utf-8
''' General tools for the Jupyter Notebook and Lab '''

from IPython.display import display
from ipywidgets import HTML, Tab, Text, Accordion, Checkbox, HBox, Output,\
                       DOMWidget, Layout, Widget, Label, VBox, Button,\
                       ToggleButton, IntSlider
from traitlets import HasTraits, List, Unicode, observe, Instance, Tuple, All
import json

# imports for async widgets
import threading
import time

# import EE
import ee
if not ee.data._initialized: ee.Initialize()

def create_accordion(dictionary):
    """ Create an Accordion output from a dict object """
    widlist = []
    ini = 0
    widget = Accordion()
    widget.selected_index = None # this will unselect all
    for key, val in dictionary.items():
        if isinstance(val, dict):
            newwidget = create_accordion(val)
            widlist.append(newwidget)
        elif isinstance(val, list):
            # tranform list to a dictionary
            dictval = {k: v for k, v in enumerate(val)}
            newwidget = create_accordion(dictval)
            widlist.append(newwidget)
        else:
            value = HTML(str(val))
            widlist.append(value)
        widget.set_title(ini, key)
        ini += 1
    widget.children = widlist
    return widget


class CheckRow(HBox):
    checkbox = Instance(Checkbox)
    widget = Instance(Widget)

    def __init__(self, widget, **kwargs):
        self.checkbox = Checkbox(indent=False,
                                layout=Layout(flex='1 1 20', width='auto'))
        self.widget = widget
        super(CheckRow, self).__init__(children=(self.checkbox, self.widget),
                                       **kwargs)
        self.layout = Layout(display='flex', flex_flow='row',
                             align_content='flex-start')

    @observe('widget')
    def _ob_wid(self, change):
        new = change['new']
        self.children = (self.checkbox, new)

    def observe_checkbox(self, handler, extra_params={}, **kwargs):
        """ set handler for the checkbox widget

        :param handler: callback function
        :type handler: function
        :param extra_params: extra parameters that can be passed to the handler
        :type extra_params: dict
        :param kwargs: parameters from traitlets.observe
        :type kwargs: dict
        """
        def proxy_handler(handler):
            def wrap(change):
                change['widget'] = self.widget
                for key, val in extra_params.items():
                    change[key] = val
                return handler(change)
            return wrap
        self.checkbox.observe(proxy_handler(handler), **kwargs)

    def observe_widget(self, handler, extra_params={}, **kwargs):
        """ set handler for the widget alongside de checkbox

        :param handler: callback function
        :type handler: function
        :param extra_params: extra parameters that can be passed to the handler
        :type extra_params: dict
        :param kwargs: parameters from traitlets.observe
        :type kwargs: dict
        """
        def proxy_handler(handler):
            def wrap(change):
                change['checkbox'] = self.checkbox
                for key, val in extra_params.items():
                    change[key] = val
                return handler(change)
            return wrap
        self.widget.observe(proxy_handler(handler), **kwargs)

class CheckAccordion(VBox):
    widgets = Tuple()

    def __init__(self, widgets, **kwargs):
        # self.widgets = widgets
        super(CheckAccordion, self).__init__(**kwargs)
        self.widgets = widgets

    @observe('widgets')
    def _on_child(self, change):
        new = change['new'] # list of any widget
        newwidgets = []
        for widget in new:
            # constract the widget
            acc = Accordion(children=(widget,))
            acc.selected_index = None # this will unselect all
            # create a CheckRow
            checkrow = CheckRow(acc)
            newwidgets.append(checkrow)
        newchildren = tuple(newwidgets)
        self.children = newchildren

    def set_title(self, index, title):
        checkrow = self.children[index]
        acc = checkrow.widget
        acc.set_title(0, title)

    def get_title(self, index):
        checkrow = self.children[index]
        acc = checkrow.widget
        return acc.get_title(0)

    def get_check(self, index):
        ''' get the state of checkbox in index '''
        checkrow = self.children[index]
        return checkrow.checkbox.value

    def set_check(self, index, state):
        ''' set the state of checkbox in index '''
        checkrow = self.children[index]
        checkrow.checkbox.value = state

    def checked_rows(self):
        ''' return a list of indexes of checked rows '''
        checked = []
        for i, checkrow in enumerate(self.children):
            state = checkrow.checkbox.value
            if state: checked.append(i)
        return checked

    def get_widget(self, index):
        ''' get the widget in index '''
        checkrow = self.children[index]
        return checkrow.widget

    def set_widget(self, index, widget):
        ''' set the widget for index '''
        checkrow = self.children[index]
        checkrow.widget.children = (widget,) # Accordion has 1 child

    def set_row(self, index, title, widget):
        ''' set values for the row '''
        self.set_title(index, title)
        self.set_widget(index, widget)

    def set_accordion_handler(self, index, handler, **kwargs):
        ''' set the handler for Accordion in index '''
        checkrow = self.children[index]
        checkrow.observe_widget(handler, names=['selected_index'], **kwargs)

    def set_checkbox_handler(self, index, handler, **kwargs):
        ''' set the handler for CheckBox in index '''
        checkrow = self.children[index]
        checkrow.observe_checkbox(handler, **kwargs)


class AssetManager(CheckAccordion):
    """ Asset Manager Widget """
    def __init__(self, path=None, **kwargs):
        # Get Assets data
        if not path:
            self.path = ee.data.getAssetRoots()[0]['id']
        else:
            self.path = path

        root_list = ee.data.getList({'id': self.path})

        # empty lists to fill with ids, types and widgets
        ids = []
        types = []
        widgets = []
        paths = []

        # first widget (headers)
        first_wid = Label('root folder')
        widgets.append(first_wid)
        types.append('root')
        ids.append(self.path)
        paths.append(self.path)

        # iterate over the list of the root
        for content in root_list:
            # get data
            id = content['id']
            ty = content['type']
            # append data to lists
            paths.append(id)
            ids.append(id.replace(self.path, ''))
            types.append(ty)

            # TODO: change this for a callback that only creates a new AssetManaget when user clicks on asset
            if ty == 'Folder' or ty == 'ImageCollection':
                wid = HTML('')
            else:
                # create widget
                # TODO: Button to get info or load to map
                wid = HTML(ty)
            widgets.append(wid)

        super(AssetManager, self).__init__(widgets=widgets, **kwargs)
        self.widgets = widgets

        # set title of header
        self.set_title(0, self.path)

        # TODO: set handler for title's checkbox: select all checkboxes

        # set titles
        for i, title in enumerate(ids):
            if i == 0: continue
            self.set_title(i, title)

        def handle_new_accordion(change):
            path = change['path']
            index = change['index']
            wid = AssetManager(path)
            self.set_widget(index, wid)

        # set handlers
        for i, path in enumerate(paths):
            self.set_accordion_handler(i, handle_new_accordion,
                                       extra_params={'path':path, 'index':i})

class TaskManager(VBox):
    def __init__(self, **kwargs):
        super(TaskManager, self).__init__(**kwargs)
        # Header
        self.checkbox = Checkbox(indent=False,
                                 layout=Layout(flex='1 1 20', width='auto'))
        self.cancel_selected = Button(description='Cancel Selected',
                                      tooltip='Cancel all selected tasks')
        self.cancel_all = Button(description='Cancell All',
                                 tooltip='Cancel all tasks')
        self.refresh = Button(description='Refresh',
                              tooltip='Refresh Tasks List')
        self.autorefresh = ToggleButton(description='auto-refresh',
                                        tooltip='click to enable/disable autorefresh')
        self.slider = IntSlider(min=1, max=10, step=1, value=5)
        self.hbox = HBox([self.checkbox, self.refresh,
                          self.cancel_selected, self.cancel_all,
                          self.autorefresh, self.slider])

        # Tabs for COMPLETED, FAILED, etc
        self.tabs = Tab()
        # Tabs index
        self.tab_index = {0: 'RUNNING',
                          1: 'COMPLETED',
                          2: 'FAILED',
                          3: 'CANCELED',
                          4: 'UNKNOWN'}

        self.taskVBox = VBox()

        self.runningVBox = VBox()
        self.completedVBox = VBox()
        self.failedVBox = VBox()
        self.canceledVBox = VBox()
        self.unknownVBox = VBox()

        self.tab_widgets_rel = {'RUNNING': self.runningVBox,
                                'COMPLETED': self.completedVBox,
                                'FAILED': self.failedVBox,
                                'CANCELED': self.canceledVBox,
                                'UNKNOWN': self.unknownVBox}

        # Create Tabs
        self.tab_widgets = []
        for key, val in self.tab_index.items():
            widget = self.tab_widgets_rel[val]
            self.tab_widgets.append(widget)
            self.tabs.children = self.tab_widgets
            self.tabs.set_title(key, val)

        ''' autorefresh
        def update_task_list(widget):
            # widget is a VBox
            tasklist = ee.data.getTaskList()
            widlist = []
            for task in tasklist:
                accordion = create_accordion(task)
                if task.has_key('description'):
                    name = '{} ({})'.format(task['description'], task['state'])
                else:
                    name = '{} ({})'.format(task['output_url'][0].split('/')[-1], task['state'])
                mainacc = Accordion(children=(accordion, ))
                mainacc.set_title(0, name)
                mainacc.selected_index = None
                wid = CheckRow(mainacc)
                #wid = CheckRow(accordion)
                widlist.append(wid)
            widget.children = tuple(widlist)

        '''
        def loop(widget):
            while True:
                self.update_task_list()(self.refresh)
                time.sleep(self.slider.value)

        # First widget
        self.update_task_list(vbox=self.runningVBox)(self.refresh)
        # self.children = (self.hbox, self.taskVBox)
        self.children = (self.hbox, self.tabs)

        # Set on_click for refresh button
        self.refresh.on_click(self.update_task_list(vbox=self.selected_tab()))
        ''' autorefresh
        thread = threading.Thread(target=loop, args=(self.taskVBox,))
        thread.start()
        '''
        # Set on_clicks
        self.cancel_all.on_click(self.cancel_all_click)
        self.cancel_selected.on_click(self.cancel_selected_click)
        # self.autorefresh

    def autorefresh_loop(self):
        pass

    def tab_handler(self, change):
        if change['name'] == 'selected_index':
            self.update_task_list()(self.refresh)

    def selected_tab(self):
        index = self.tabs.selected_index
        tab_name = self.tab_index[index]
        return self.tab_widgets_rel[tab_name]

    def update_task_list(self, **kwargs):
        def wrap(button):
            self.selected_tab().children = (HTML('Loading...'),)
            try:
                tasklist = ee.data.getTaskList()
                # empty lists
                running_list = []
                completed_list = []
                failed_list = []
                canceled_list = []
                unknown_list = []
                all_list = {'RUNNING': running_list, 'COMPLETED': completed_list,
                            'FAILED': failed_list, 'CANCELED': canceled_list,
                            'UNKNOWN': unknown_list}
                for task in tasklist:
                    state = task['state']
                    accordion = create_accordion(task)
                    if task['state'] == 'COMPLETED':
                        start = int(task['start_timestamp_ms'])
                        end = int(task['creation_timestamp_ms'])
                        seconds = float((start-end))/1000
                        name = '{} ({} sec)'.format(task['output_url'][0].split('/')[-1],
                                                         seconds)
                    else:
                        name = '{}'.format(task['description'])
                    # Accordion for CheckRow widget
                    mainacc = Accordion(children=(accordion, ))
                    mainacc.set_title(0, name)
                    mainacc.selected_index = None
                    # CheckRow
                    wid = CheckRow(mainacc)
                    # Append widget to the CORRECT list
                    all_list[state].append(wid)
                # Assign Children
                self.runningVBox.children = tuple(running_list)
                self.completedVBox.children = tuple(completed_list)
                self.failedVBox.children = tuple(failed_list)
                self.canceledVBox.children = tuple(canceled_list)
                self.unknownVBox.children = tuple(unknown_list)
            except Exception as e:
                self.selected_tab().children = (HTML(str(e)),)
        return wrap

    def get_selected(self):
        """ Get selected Tasks

        :return: a list of the selected indexes
        """
        selected = []
        children = self.selected_tab().children
        for i, child in enumerate(children):
            # checkrow = child.children[0] # child is an accordion
            state = child.checkbox.value
            if state: selected.append(i)
        return selected

    def get_taskid(self, index):
        # Get selected Tab
        selected_wid = self.selected_tab() # VBox
        # Children of the Tab's VBox
        children = selected_wid.children
        # Get CheckRow that corresponds to the passed index
        checkrow = children[index]
        # Get main accordion
        mainacc = checkrow.widget
        # Get details accordion
        selectedacc = mainacc.children[0]
        for n, child in enumerate(selectedacc.children):
            title = selectedacc.get_title(n)
            if title == 'id':
                return child.value

    def get_selected_taskid(self):
        selected = self.get_selected()
        selected_wid = self.selected_tab() # VBox
        children = selected_wid.children
        taskid_list = []
        for select in selected:
            '''
            checkrow = children[select]
            mainacc = checkrow.widget
            selectedacc = mainacc.children[0]
            for n, child in enumerate(selectedacc.children):
                title = selectedacc.get_title(n)
                if title == 'id':
                    taskid_list.append(child.value)
            '''
            taskid = self.get_taskid(select)
            taskid_list.append(taskid)

        return taskid_list

    def cancel_selected_click(self, button):
        selected = self.get_selected_taskid()
        for taskid in selected:
            try:
                ee.data.cancelTask(taskid)
            except:
                continue
        self.update_task_list()(self.refresh)

    def cancel_all_click(self, button):
        selected_wid = self.selected_tab() # VBox
        children = selected_wid.children
        for n, child in enumerate(children):
            taskid = self.get_taskid(n)
            try:
                ee.data.cancelTask(taskid)
            except:
                continue
        self.update_task_list()(self.refresh)



# coding=utf-8

''' This module is designed to use ONLY in the Jupyter Notebook. It is
 inspired on Tyler Erickson's contribution on
https://github.com/gee-community/ee-jupyter-contrib/blob/master/examples/getting-started/display-interactive-map.ipynb'''

import ipyleaflet
from ipywidgets import HTML, Tab, Accordion, HBox, SelectMultiple, Select,\
                       Button, VBox, RadioButtons, Dropdown, Layout, \
                       FloatRangeSlider, Label
from IPython.display import display
from traitlets import Dict, observe, HasTraits, Unicode, Int, Float, Instance
import ee
if not ee.data._initialized: ee.Initialize()
from collections import OrderedDict
from . import tools
from .maptool import get_default_vis, inverse_coordinates, get_data,\
                     get_image_tile, get_geojson_tile, get_bounds, get_zoom,\
                     create_html, feature_properties_output
import maptool
from . import ipytools
import threading
from copy import copy
import traceback
import sys


class Map(ipyleaflet.Map):
    tab_children_dict = Dict()
    EELayers = Dict()
    def __init__(self, tabs=('Inspector', 'Layers', 'Assets', 'Tasks'),
                 **kwargs):
        # Change defaults
        kwargs.setdefault('center', [0, 0])
        kwargs.setdefault('zoom', 2)
        kwargs.setdefault('scroll_wheel_zoom', True)
        super(Map, self).__init__(**kwargs)
        self.is_shown = False

        # Correct base layer name
        baselayer = self.layers[0]
        baselayer.name = 'OpenStreetMap'
        self.layers = (baselayer,)

        # Dictionary of map's handlers
        self.handlers = {}

        # Dictonary to hold tab's widgets
        # (tab's name:widget)
        self.tab_names = []
        self.tab_children = []
        self.tab_children_dict = OrderedDict(zip(self.tab_names,
                                                 self.tab_children))

        # TABS
        # Tab widget
        self.tab_widget = Tab()
        # Handler for Tab
        self.tab_widget.observe(self.handle_change_tab)

        self.tabs = tabs
        if len(tabs) > 0:
            # TODO: create widgets only if are in tuple
            # Inspector Widget (Accordion)
            self.inspector_wid = CustomInspector()
            self.inspector_wid.main.selected_index = None # this will unselect all

            # Task Manager Widget
            task_manager = ipytools.TaskManager()

            # Asset Manager Widget
            asset_manager = ipytools.AssetManager(self)

            # Layers
            self.layers_widget = LayersWidget(map=self)

            widgets = {'Inspector': self.inspector_wid,
                       'Layers': self.layers_widget,
                       'Assets': asset_manager,
                       'Tasks': task_manager,
                       }
            handlers = {'Inspector': self.handle_inspector,
                        'Layers': None,
                        'Assets': None,
                        'Tasks': None,
                        }

            # Add tabs and handlers
            for tab in tabs:
                if tab in widgets.keys():
                    widget = widgets[tab]
                    handler = handlers[tab]
                    self.addTab(tab, handler, widget)
                else:
                    raise ValueError('Tab {} is not recognized. Choose one of {}'.format(tab, widgets.keys()))

            # First handler: Inspector
            self.on_interaction(self.handlers[tabs[0]])

        # As I cannot create a Geometry with a GeoJSON string I do a workaround
        self.draw_types = {'Polygon': ee.Geometry.Polygon,
                           'Point': ee.Geometry.Point,
                           'LineString': ee.Geometry.LineString,
                           }
        # create EELayers
        self.EELayers = OrderedDict()

    def _add_EELayer(self, name, data):
        ''' add a pair of name, data to EELayers '''
        copyEELayers = copy(self.EELayers)
        copyEELayers[name] = data
        self.EELayers = copyEELayers

    def _remove_EELayer(self, name):
        ''' remove layer from EELayers '''
        copyEELayers = copy(self.EELayers)
        if name in copyEELayers:
            copyEELayers.pop(name)
        self.EELayers = copyEELayers

    def move(self, layer_name, direction='up'):
        ''' Move one step up a layer '''
        names = self.EELayers.keys()
        values = self.EELayers.values()

        if direction == 'up':
            dir = 1
        elif direction == 'down':
            dir = -1
        else:
            dir = 0

        if layer_name in names:  # if layer exists
            # index and value of layer to move
            i = names.index(layer_name)
            condition = (i < len(names)-1) if dir == 1 else (i > 0)
            if condition:  # if layer is not in the edge
                ival = values[i]
                # new index for layer
                newi = i+dir
                # get index and value that already exist in the new index
                iname_before = names[newi]
                ival_before = values[newi]
                # Change order
                # set layer and value in the new index
                names[newi] = layer_name
                values[newi] = ival
                # set replaced layer and its value in the index of moving layer
                names[i] = iname_before
                values[i] = ival_before

                newlayers = OrderedDict(zip(names, values))
                self.EELayers = newlayers

    @observe('EELayers')
    def _ob_EELayers(self, change):
        new = change['new']
        proxy_layers = [self.layers[0]]

        for val in new.values():
            layer = val['layer']
            proxy_layers.append(layer)

        self.layers = tuple(proxy_layers)

        # UPDATE INSPECTOR
        # Clear options
        self.inspector_wid.selector.options = {}
        # Add layer to the Inspector Widget
        self.inspector_wid.selector.options = new # self.EELayers

        # UPDATE LAYERS WIDGET
        # update Layers Widget
        self.layers_widget.selector.options = {}
        self.layers_widget.selector.options = new # self.EELayers

    @property
    def added_images(self):
        return sum(
            [1 for val in self.EELayers.values() if val['type'] == 'Image'])

    @property
    def added_geometries(self):
        return sum(
            [1 for val in self.EELayers.values() if val['type'] == 'Geometry'])

    def task_widget(self):
        with self.tasksWid:
            while True:
                list = ee.data.getTaskList()

    def show(self, tabs=True, layer_control=True, draw_control=False):
        """ Show the Map on the Notebook """
        if not self.is_shown:
            if layer_control:
                # Layers Control
                lc = ipyleaflet.LayersControl()
                self.add_control(lc)
            if draw_control:
                # Draw Control
                dc = ipyleaflet.DrawControl(# edit=False,
                                            # marker={'shapeOptions': {}}
                                            )
                dc.on_draw(self.handle_draw)
                self.add_control(dc)

            if tabs:
                display(self, self.tab_widget)
            else:
                display(self)
        else:
            # if len(tabs) > 0:
            if tabs:
                display(self, self.tab_widget)
            else:
                display(self)

        self.is_shown = True

    def show_tab(self, name):
        """ Show only a Tab Widget by calling its name. This is useful mainly
        in Jupyter Lab where you can see outputs in different tab_widget

        :param name: the name of the tab to show
        :type name: str
        """
        try:
            widget = self.tab_children_dict[name]
            display(widget)
        except:
            print('Tab not found')

    def addImage(self, image, visParams=None, name=None, show=True,
                 opacity=None, replace=True):
        """ Add an ee.Image to the Map

        :param image: Image to add to Map
        :type image: ee.Image
        :param visParams: visualization parameters. Can have the
            following arguments: bands, min, max.
        :type visParams: dict
        :param name: name for the layer
        :type name: str
        :return: the name of the added layer
        :rtype: str
        """
        # Check if layer exists
        if name in self.EELayers.keys():
            if not replace:
                msg = "Image with name '{}' exists already, please choose " \
                      "another name"
                print(msg.format(name))
                return
            else:
                # Get URL, attribution & vis params
                params = get_image_tile(image, visParams, show, opacity)

                # Remove Layer
                self.removeLayer(name)
        else:
            # Get URL, attribution & vis params
            params = get_image_tile(image, visParams, show, opacity)

        layer = ipyleaflet.TileLayer(url=params['url'],
                                     attribution=params['attribution'],
                                     name=name)

        EELayer = {'type': 'Image',
                   'object': image,
                   'visParams': params['visParams'],
                   'layer': layer}

        # self._add_EELayer(name, EELayer)
        # return name
        return EELayer

    def addMarker(self, marker, visParams=None, name=None, show=True,
                  opacity=None, replace=True,
                  inspect={'data':None, 'reducer':None, 'scale':None}):
        ''' General method to add Geometries, Features or FeatureCollections
        as Markers '''

        if isinstance(marker, ee.Geometry):
            self.addGeometry(marker, visParams, name, show, opacity, replace,
                             inspect)

        elif isinstance(marker, ee.Feature):
            self.addFeature(marker, visParams, name, show, opacity, replace,
                             inspect)

        elif isinstance(marker, ee.FeatureCollection):
            geometry = marker.geometry()
            self.addGeometry(marker, visParams, name, show, opacity, replace,
                             inspect)

    def addFeature(self, feature, visParams=None, name=None, show=True,
                    opacity=None, replace=True,
                    inspect={'data':None, 'reducer':None, 'scale':None}):
        """ Add a Feature to the Map

        :param feature: the Feature to add to Map
        :type feature: ee.Feature
        :param visParams:
        :type visParams: dict
        :param name: name for the layer
        :type name: str
        :param inspect: when adding a geometry or a feature you can pop up data
            from a desired layer. Params are:
            :data: the EEObject where to get the data from
            :reducer: the reducer to use
            :scale: the scale to reduce
        :type inspect: dict
        :return: the name of the added layer
        :rtype: str
        """
        thename = name if name else 'Feature {}'.format(self.added_geometries)

        # Check if layer exists
        if thename in self.EELayers.keys():
            if not replace:
                print("Layer with name '{}' exists already, please choose another name".format(thename))
                return
            else:
                self.removeLayer(thename)

        params = get_geojson_tile(feature, thename, inspect)
        layer = ipyleaflet.GeoJSON(data=params['geojson'],
                                   name=thename,
                                   popup=HTML(params['pop']))

        self._add_EELayer(thename, {'type': 'Feature',
                                    'object': feature,
                                    'visParams': None,
                                    'layer': layer})
        return thename

    def addGeometry(self, geometry, visParams=None, name=None, show=True,
                    opacity=None, replace=True,
                    inspect={'data':None, 'reducer':None, 'scale':None}):
        """ Add a Geometry to the Map

        :param geometry: the Geometry to add to Map
        :type geometry: ee.Geometry
        :param visParams:
        :type visParams: dict
        :param name: name for the layer
        :type name: str
        :param inspect: when adding a geometry or a feature you can pop up data
            from a desired layer. Params are:
            :data: the EEObject where to get the data from
            :reducer: the reducer to use
            :scale: the scale to reduce
        :type inspect: dict
        :return: the name of the added layer
        :rtype: str
        """
        thename = name if name else 'Geometry {}'.format(self.added_geometries)

        # Check if layer exists
        if thename in self.EELayers.keys():
            if not replace:
                print("Layer with name '{}' exists already, please choose another name".format(thename))
                return
            else:
                self.removeLayer(thename)

        params = get_geojson_tile(geometry, thename, inspect)
        layer = ipyleaflet.GeoJSON(data=params['geojson'],
                                   name=thename,
                                   popup=HTML(params['pop']))

        self._add_EELayer(thename, {'type': 'Geometry',
                                      'object': geometry,
                                      'visParams':None,
                                      'layer': layer})
        return thename

    def addFeatureLayer(self, feature, visParams=None, name=None, show=True,
                        opacity=None, replace=True):
        ''' Paint a Feature on the map, but the layer underneath is the
        actual added Feature '''

        visParams = visParams if visParams else {}

        if isinstance(feature, ee.Feature):
            ty = 'Feature'
        elif isinstance(feature, ee.FeatureCollection):
            ty = 'FeatureCollection'
        else:
            print('The object is not a Feature or FeatureCollection')
            return

        fill_color = visParams.get('fill_color', None)

        if 'outline_color' in visParams:
            out_color = visParams['outline_color']
        elif 'border_color' in visParams:
            out_color = visParams['border_color']
        else:
            out_color = 'black'

        outline = visParams.get('outline', 2)

        proxy_layer = maptool.paint(feature, out_color, fill_color, outline)

        thename = name if name else '{} {}'.format(ty, self.added_geometries)

        img_params = {'bands':['vis-red', 'vis-green', 'vis-blue'],
                      'min': 0, 'max':255}

        # Check if layer exists
        if thename in self.EELayers.keys():
            if not replace:
                print("{} with name '{}' exists already, please choose another name".format(ty, thename))
                return
            else:
                # Get URL, attribution & vis params
                params = get_image_tile(proxy_layer, img_params, show, opacity)

                # Remove Layer
                self.removeLayer(thename)
        else:
            # Get URL, attribution & vis params
            params = get_image_tile(proxy_layer, img_params, show, opacity)

        layer = ipyleaflet.TileLayer(url=params['url'],
                                     attribution=params['attribution'],
                                     name=thename)

        self._add_EELayer(thename, {'type': ty,
                                    'object': feature,
                                    'visParams': visParams,
                                    'layer': layer})
        return thename

    def addMosaic(self, collection, visParams=None, name=None, show=False,
                  opacity=None, replace=True):
        ''' Add an ImageCollection to EELayer and its mosaic to the Map.
        When using the inspector over this layer, it will print all values from
        the collection '''
        proxy = ee.ImageCollection(collection).sort('system:time_start')
        mosaic = ee.Image(proxy.mosaic())

        EELayer = self.addImage(mosaic, visParams, name, show, opacity, replace)
        # modify EELayer
        EELayer['type'] = 'ImageCollection'
        EELayer['object'] = ee.ImageCollection(collection)
        return EELayer

    def addImageCollection(self, collection, visParams=None, nametags=['id'],
                           show=False, opacity=None):
        """ Add every Image of an ImageCollection to the Map

        :param collection: the ImageCollection
        :type collection: ee.ImageCollection
        :param visParams: visualization parameter for each image. See `addImage`
        :type visParams: dict
        :param nametags: tags that will be the name for each image. It must be
            a list in which each element is a string. Each string can be any
            Image property, or one of the following:
            - system_date: the name will be the date of each Image
            - id: the name will be the ID of each Image (Default)
        :type nametags: list
        :param show: If True, adds and shows the Image, otherwise only add it
        :type show: bool
        """
        size = collection.size().getInfo()
        collist = collection.toList(size)
        separation = ' '
        for inx in range(size):
            img = ee.Image(collist.get(inx))
            name = ''
            properties = img.propertyNames().getInfo()
            for nametag in nametags:
                if nametag == 'id':
                    newname = img.id().getInfo()
                elif nametag == 'system_date':
                    newname = ee.Date(img.date()).format('YYYY-MM-dd').getInfo()
                elif nametag in properties:
                    newname = "{}:{}{}".format(nametag, img.get(nametag).getInfo(), separation)
                else:
                    newname = img.id().getInfo()

                name += newname
            self.addLayer(img, visParams, str(name), show, opacity)

    def addLayer(self, eeObject, visParams=None, name=None, show=True,
                 opacity=None, replace=True, **kwargs):
        """ Adds a given EE object to the map as a layer.

        :param eeObject: Earth Engine object to add to map
        :type eeObject: ee.Image || ee.Geometry || ee.Feature
        :param replace: if True, if there is a layer with the same name, this
            replace that layer.
        :type replace: bool

        For ee.Image and ee.ImageCollection see `addImage`
        for ee.Geometry and ee.Feature see `addGeometry`
        """
        visParams = visParams if visParams else {}

        # CASE: ee.Image
        if isinstance(eeObject, ee.Image):
            image_name = name if name else 'Image {}'.format(self.added_images)
            EELayer = self.addImage(eeObject, visParams=visParams,
                                    name=image_name, show=show,
                                    opacity=opacity, replace=replace)

            self._add_EELayer(image_name, EELayer)
            added_layer = EELayer

        # CASE: ee.Geometry
        elif isinstance(eeObject, ee.Geometry):
            geom = eeObject if isinstance(eeObject, ee.Geometry) else eeObject.geometry()
            kw = {'visParams':visParams, 'name':name, 'show':show, 'opacity':opacity}
            if kwargs.get('inspect'): kw.setdefault('inspect', kwargs.get('inspect'))
            added_layer = self.addGeometry(geom, replace=replace, **kw)

        # CASE: ee.Feature & ee.FeatureCollection
        elif isinstance(eeObject, ee.Feature) or isinstance(eeObject, ee.FeatureCollection):
            feat = eeObject
            kw = {'visParams':visParams, 'name':name, 'show':show, 'opacity':opacity}
            added_layer = self.addFeatureLayer(feat, replace=replace, **kw)

        # CASE: ee.ImageCollection
        elif isinstance(eeObject, ee.ImageCollection):
            '''
            proxy = eeObject.sort('system:time_start')
            mosaic = ee.Image(proxy.mosaic())
            added_layer = self.addImage(mosaic, visParams=visParams, name=thename,
                                        show=show, opacity=opacity, replace=replace)
            '''
            thename = name if name else 'ImageCollection {}'.format(self.added_images)
            EELayer = self.addMosaic(eeObject, visParams, thename, show,
                                     opacity, replace)
            self._add_EELayer(thename, EELayer)

            added_layer = EELayer

        else:
            added_layer = None
            print("`addLayer` doesn't support adding {} objects to the map".format(type(eeObject)))

        # return added_layer

    def removeLayer(self, name):
        """ Remove a layer by its name """
        if name in self.EELayers.keys():
            self._remove_EELayer(name)
        else:
            print('Layer {} is not present in the map'.format(name))
            return

    def getLayer(self, name):
        """ Get a layer by its name

        :param name: the name of the layer
        :type name: str
        :return: The complete EELayer which is a dict of

            :type: the type of the layer
            :object: the EE Object associated with the layer
            :visParams: the visualization parameters of the layer
            :layer: the TileLayer added to the Map (ipyleaflet.Map)

        :rtype: dict
        """
        if name in self.EELayers:
            layer = self.EELayers[name]
            return layer
        else:
            print('Layer {} is not present in the map'.format(name))
            return

    def getObject(self, name):
        ''' Get the EE Object from a layer's name '''
        obj = self.getLayer(name)['object']
        return obj

    def getVisParams(self, name):
        ''' Get the Visualization Parameters from a layer's name '''
        vis = self.getLayer(name)['visParams']
        return vis

    def centerObject(self, eeObject, zoom=None, method=1):
        """ Center an eeObject

        :param eeObject:
        :param zoom:
        :param method: experimetal methods to estimate zoom for fitting bounds
            Currently: 1 or 2
        :type: int
        """
        bounds = get_bounds(eeObject)
        if bounds:
            try:
                inverse = inverse_coordinates(bounds)
                centroid = ee.Geometry.Polygon(inverse)\
                             .centroid().getInfo()['coordinates']
            except:
                centroid = [0, 0]

            self.center = inverse_coordinates(centroid)
            if zoom:
                self.zoom = zoom
            else:
                self.zoom = get_zoom(bounds, method)

    def getCenter(self):
        """ Returns the coordinates at the center of the map.

        No arguments.
        Returns: Geometry.Point

        :return:
        """
        center = self.center
        coords = inverse_coordinates(center)
        return ee.Geometry.Point(coords)

    def getBounds(self, asGeoJSON=True):
        """ Returns the bounds of the current map view, as a list in the
        format [west, south, east, north] in degrees.

        Arguments:
        asGeoJSON (Boolean, optional):
        If true, returns map bounds as GeoJSON.

        Returns: GeoJSONGeometry|List<Number>|String
        """
        bounds = inverse_coordinates(self.bounds)
        if asGeoJSON:
            return ee.Geometry.Rectangle(bounds)
        else:
            return bounds

    def _update_tab_children(self):
        """ Update Tab children from tab_children_dict """
        # Set tab_widget children
        self.tab_widget.children = self.tab_children_dict.values()
        # Set tab_widget names
        for i, name in enumerate(self.tab_children_dict.keys()):
            self.tab_widget.set_title(i, name)

    def addTab(self, name, handler=None, widget=None):
        """ Add a Tab to the Panel. The handler is for the Map

        :param name: name for the new tab
        :type name: str
        :param handler: handle function for the new tab. Arguments of the
            function are:

            :type: the type of the event (click, mouseover, etc..)
            :coordinates: coordinates where the event occured [lon, lat]
            :widget: the widget inside the Tab
            :map: the Map instance

        :param widget: widget inside the Tab. Defaults to HTML('')
        :type widget: ipywidgets.Widget
        """
        # Widget
        wid = widget if widget else HTML('')
        # Get tab's children as a list
        # tab_children = list(self.tab_widget.children)
        tab_children = self.tab_children_dict.values()
        # Get a list of tab's titles
        # titles = [self.tab_widget.get_title(i) for i, child in enumerate(tab_children)]
        titles = self.tab_children_dict.keys()
        # Check if tab already exists
        if name not in titles:
            ntabs = len(tab_children)

            # UPDATE DICTS
            # Add widget as a new children
            self.tab_children_dict[name] = wid
            # Set the handler for the new tab
            if handler:
                def proxy_handler(f):
                    def wrap(**kwargs):
                        # Add widget to handler arguments
                        kwargs['widget'] = self.tab_children_dict[name]
                        coords = kwargs['coordinates']
                        kwargs['coordinates'] = inverse_coordinates(coords)
                        kwargs['map'] = self
                        return f(**kwargs)
                    return wrap
                self.handlers[name] = proxy_handler(handler)
            else:
                self.handlers[name] = handler

            # Update tab children
            self._update_tab_children()
        else:
            print('Tab {} already exists, please choose another name'.format(name))

    def handle_change_tab(self, change):
        """ Handle function to trigger when tab changes """
        # Remove all handlers
        if change['name'] == 'selected_index':
            old = change['old']
            new = change['new']
            old_name = self.tab_widget.get_title(old)
            new_name = self.tab_widget.get_title(new)
            # Remove all handlers
            for handl in self.handlers.values():
                self.on_interaction(handl, True)
            # Set new handler if not None
            if new_name in self.handlers.keys():
                handler = self.handlers[new_name]
                if handler:
                    self.on_interaction(handler)

    def handle_inspector(self, **change):
        """ Handle function for the Inspector Widget """
        # Get click coordinates
        coords = change['coordinates']

        event = change['type'] # event type
        if event == 'click':  # If the user clicked
            # create a point where the user clicked
            point = ee.Geometry.Point(coords)

            # Get widget
            thewidget = change['widget'].main  # Accordion

            # First Accordion row text (name)
            first = 'Point {} at {} zoom'.format(coords, self.zoom)
            namelist = [first]
            wids4acc = [HTML('')] # first row has no content

            # Get only Selected Layers in the Inspector Selector
            selected_layers = dict(zip(self.inspector_wid.selector.label,
                                       self.inspector_wid.selector.value))

            length = len(selected_layers.keys())
            i = 1

            for name, obj in selected_layers.items(): # for every added layer
                # Clear children // Loading
                thewidget.children = [HTML('wait a second please..')]
                thewidget.set_title(0, 'Loading {} of {}...'.format(i, length))
                i += 1

                # Image
                if obj['type'] == 'Image':
                    # Get the image's values
                    try:
                        image = obj['object']
                        values = tools.get_value(image, point, 10, 'client')
                        values = tools.sort_dict(values)
                        # Create the content
                        img_html = ''
                        for band, value in values.items():
                            img_html += '<b>{}</b>: {}</br>'.format(band,
                                                                    value)
                        wid = HTML(img_html)
                        # append widget to list of widgets
                        wids4acc.append(wid)
                        namelist.append(name)
                    except Exception as e:
                        # wid = HTML(str(e).replace('<','{').replace('>','}'))
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        trace = traceback.format_exception(exc_type, exc_value,
                                                           exc_traceback)
                        wid = ErrorAccordion(e, trace)
                        wids4acc.append(wid)
                        namelist.append('ERROR at layer {}'.format(name))

                # ImageCollection
                if obj['type'] == 'ImageCollection':
                    # Get the values from all images
                    try:
                        collection = obj['object']
                        values = tools.get_values(collection, point, 10,
                                                  'client')

                        # header
                        allbands = [val.keys() for bands, val in values.items()]
                        bands = []
                        for bandlist in allbands:
                            for band in bandlist:
                                if band not in bands:
                                    bands.append(band)

                        header = ['image']+bands

                        # rows
                        rows = []
                        for imgid, val in values.items():
                            row = ['']*len(header)
                            row[0] = str(imgid)
                            for bandname, bandvalue in val.items():
                                pos = header.index(bandname) if bandname in header else None
                                if pos:
                                    row[pos] = str(bandvalue)
                            rows.append(row)

                        # Create the content
                        html = maptool.create_html_table(header, rows)
                        wid = HTML(html)
                        # append widget to list of widgets
                        wids4acc.append(wid)
                        namelist.append(name)
                    except Exception as e:
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        trace = traceback.format_exception(exc_type, exc_value,
                                                           exc_traceback)
                        wid = ErrorAccordion(e, trace)
                        wids4acc.append(wid)
                        namelist.append('ERROR at layer {}'.format(name))

                # Features
                if obj['type'] == 'Feature':
                    try:
                        feat = obj['object']
                        feat_geom = feat.geometry()
                        if feat_geom.contains(point).getInfo():
                            info = feature_properties_output(feat)
                            wid = HTML(info)
                            # append widget to list of widgets
                            wids4acc.append(wid)
                            namelist.append(name)
                    except Exception as e:
                        # wid = HTML(str(e).replace('<','{').replace('>','}'))
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        trace = traceback.format_exception(exc_type, exc_value,
                                                           exc_traceback)
                        wid = ErrorAccordion(e, trace)
                        wids4acc.append(wid)
                        namelist.append('ERROR at layer {}'.format(name))

                # FeatureCollections
                if obj['type'] == 'FeatureCollection':
                    try:
                        fc = obj['object']
                        filtered = fc.filterBounds(point)
                        if filtered.size().getInfo() > 0:
                            feat = ee.Feature(filtered.first())
                            info = feature_properties_output(feat)
                            wid = HTML(info)
                            # append widget to list of widgets
                            wids4acc.append(wid)
                            namelist.append(name)
                    except Exception as e:
                        wid = HTML(str(e).replace('<','{').replace('>','}'))
                        wids4acc.append(wid)
                        namelist.append('ERROR at layer {}'.format(name))

            # Set children and children's name of inspector widget
            thewidget.children = wids4acc
            for i, n in enumerate(namelist):
                thewidget.set_title(i, n)

    def handle_object_inspector(self, **change):
        """ Handle function for the Object Inspector Widget

        DEPRECATED
        """
        event = change['type'] # event type
        thewidget = change['widget']
        if event == 'click':  # If the user clicked
            # Clear children // Loading
            thewidget.children = [HTML('wait a second please..')]
            thewidget.set_title(0, 'Loading...')

            widgets = []
            i = 0

            for name, obj in self.EELayers.items(): # for every added layer
                the_object = obj['object']
                try:
                    properties = the_object.getInfo()
                    wid = ipytools.create_accordion(properties) # Accordion
                    wid.selected_index = None # this will unselect all
                except Exception as e:
                    wid = HTML(str(e))
                widgets.append(wid)
                thewidget.set_title(i, name)
                i += 1

            thewidget.children = widgets

    def handle_draw(self, dc_widget, action, geo_json):
        """ Handles drawings """
        ty = geo_json['geometry']['type']
        coords = geo_json['geometry']['coordinates']
        geom = self.draw_types[ty](coords)
        if action == 'created':
            self.addGeometry(geom)
        elif action == 'deleted':
            for key, val in self.EELayers.items():
                if geom == val:
                    self.removeLayer(key)


class CustomInspector(HBox):
    def __init__(self, **kwargs):
        desc = 'Select one or more layers'
        super(CustomInspector, self).__init__(description=desc, **kwargs)
        self.selector = SelectMultiple()
        self.main = Accordion()
        self.children = [self.selector, self.main]


class ErrorAccordion(Accordion):
    def __init__(self, error, traceback, **kwargs):
        super(ErrorAccordion, self).__init__(**kwargs)
        self.error = '{}'.format(error).replace('<','{').replace('>','}')

        newtraceback = ''
        for trace in traceback[1:]:
            newtraceback += '{}'.format(trace).replace('<','{').replace('>','}')
            newtraceback += '</br>'

        self.traceback = newtraceback

        self.errorWid = HTML(self.error)

        self.traceWid = HTML(self.traceback)

        self.children = (self.errorWid, self.traceWid)
        self.set_title(0, 'ERROR')
        self.set_title(1, 'TRACEBACK')

class LayersWidget(ipytools.RealBox):
    def __init__(self, map=None, **kwargs):
        super(LayersWidget, self).__init__(**kwargs)
        self.map = map
        self.selector = Select()

        # define init EELayer
        self.EELayer = None

        # Buttons
        self.center = Button(description='Center')
        self.center.on_click(self.on_click_center)

        self.remove = Button(description='Remove')
        self.remove.on_click(self.on_click_remove)

        self.show_prop = Button(description='Show Object')
        self.show_prop.on_click(self.on_click_show_object)

        self.vis = Button(description='Visualization')
        self.vis.on_click(self.on_click_vis)

        self.move_up = Button(description='Move up')
        self.move_up.on_click(self.on_up)

        self.move_down = Button(description='Move down')
        self.move_down.on_click(self.on_down)

        # Buttons Group 1
        self.group1 = VBox([self.center, self.remove,
                            self.vis, self.show_prop])

        # Buttons Group 2
        self.group2 = VBox([self.move_up, self.move_down])

        # self.children = [self.selector, self.group1]
        self.items = [[self.selector, self.group1, self.group2]]

        self.selector.observe(self.handle_selection, names='value')

    def on_up(self, button=None):
        if self.EELayer:
            self.map.move(self.layer.name, 'up')

    def on_down(self, button=None):
        if self.EELayer:
            self.map.move(self.layer.name, 'down')

    def handle_selection(self, change):
        new = change['new']
        self.EELayer = new

        # set original display
        self.items = [[self.selector, self.group1, self.group2]]

        if new:
            self.layer = new['layer']
            self.obj = new['object']
            self.ty = new['type']
            self.vis = new['visParams']

    def on_click_show_object(self, button=None):
        if self.EELayer:
            loading = HTML('Loading <b>{}</b>...'.format(self.layer.name))
            widget = VBox([loading])
            # widget = ipytools.create_object_output(self.obj)
            thread = threading.Thread(target=ipytools.create_async_output,
                                      args=(self.obj, widget))
            self.items = [[self.selector, self.group1],
                          [widget]]
            thread.start()

    def on_click_center(self, button=None):
        if self.EELayer:
            self.map.centerObject(self.obj)

    def on_click_remove(self, button=None):
        if self.EELayer:
            self.map.removeLayer(self.layer.name)

    def on_click_vis(self, button=None):
        if self.EELayer:
            # options
            selector = self.selector
            group1 = self.group1

            # map
            map = self.map
            layer_name = self.layer.name
            image = self.obj

            # Image Bands
            try:
                info = self.obj.getInfo()
            except Exception as e:
                self.items = [[self.selector, self.group1],
                              [HTML(str(e))]]
                return

            # IMAGES
            if self.ty == 'Image':
                ### image data ###
                bands = info['bands']
                imbands = [band['id'] for band in bands]
                bands_type = [band['data_type']['precision'] for band in bands]
                bands_min = []
                bands_max = []
                # as float bands don't hava an specific range, reduce region to get the
                # real range
                if 'float' in bands_type:
                    try:
                        minmax = image.reduceRegion(ee.Reducer.minMax())
                        for band in bands:
                            bandname = band['id']
                            try:
                                tmin = minmax.get('{}_min'.format(bandname)).getInfo() # 0
                                tmax = minmax.get('{}_max'.format(bandname)).getInfo() # 1
                            except:
                                tmin = 0
                                tmax = 1
                            bands_min.append(tmin)
                            bands_max.append(tmax)
                    except:
                        for band in bands:
                            dt = band['data_type']
                            try:
                                tmin = dt['min']
                                tmax = dt['max']
                            except:
                                tmin = 0
                                tmax = 1
                            bands_min.append(tmin)
                            bands_max.append(tmax)
                else:
                    for band in bands:
                        dt = band['data_type']
                        try:
                            tmin = dt['min']
                            tmax = dt['max']
                        except:
                            tmin = 0
                            tmax = 1
                        bands_min.append(tmin)
                        bands_max.append(tmax)


                # dict of {band: min} and {band:max}
                min_dict = dict(zip(imbands, bands_min))
                max_dict = dict(zip(imbands, bands_max))
                ######

                # Layer data
                layer_data = self.map.EELayers[layer_name]
                visParams = layer_data['visParams']

                # vis bands
                visBands = visParams['bands'].split(',')

                # vis min
                visMin = visParams['min']
                if isinstance(visMin, str):
                    visMin = [float(vis) for vis in visMin.split(',')]
                else:
                    visMin = [visMin]

                # vis max
                visMax = visParams['max']
                if isinstance(visMax, str):
                    visMax = [float(vis) for vis in visMax.split(',')]
                else:
                    visMax = [visMax]

                # dropdown handler
                def handle_dropdown(band_slider):
                    def wrap(change):
                        new = change['new']
                        band_slider.min = min_dict[new]
                        band_slider.max = max_dict[new]
                    return wrap

                def slider_1band(float=False, name='band'):
                    ''' Create the widget for one band '''
                    # get params to set in slider and dropdown
                    vismin = visMin[0]
                    vismax = visMax[0]
                    band = visBands[0]

                    drop = Dropdown(description=name, options=imbands, value=band)

                    if float:
                        slider = ipytools.FloatBandWidget(min=min_dict[drop.value],
                                                          max=max_dict[drop.value])
                    else:
                        slider = FloatRangeSlider(min=min_dict[drop.value],
                                                  max=max_dict[drop.value],
                                                  value=[vismin, vismax],
                                                  step=0.01)
                    # set handler
                    drop.observe(handle_dropdown(slider), names=['value'])

                    # widget for band selector + slider
                    band_slider = HBox([drop, slider])
                    # return VBox([band_slider], layout=Layout(width='500px'))
                    return band_slider

                def slider_3bands(float=False):
                    ''' Create the widget for one band '''
                    # get params to set in slider and dropdown
                    if len(visMin) == 1:
                        visminR = visminG = visminB = visMin[0]
                    else:
                        visminR = visMin[0]
                        visminG = visMin[1]
                        visminB = visMin[2]

                    if len(visMax) == 1:
                        vismaxR = vismaxG = vismaxB = visMax[0]
                    else:
                        vismaxR = visMax[0]
                        vismaxG = visMax[1]
                        vismaxB = visMax[2]

                    if len(visBands) == 1:
                        visbandR = visbandG = visbandB = visBands[0]
                    else:
                        visbandR = visBands[0]
                        visbandG = visBands[1]
                        visbandB = visBands[2]

                    drop = Dropdown(description='red', options=imbands, value=visbandR)
                    drop2 = Dropdown(description='green', options=imbands, value=visbandG)
                    drop3 = Dropdown(description='blue', options=imbands, value=visbandB)
                    slider = FloatRangeSlider(min=min_dict[drop.value],
                                              max=max_dict[drop.value],
                                              value=[visminR, vismaxR],
                                              step=0.01)
                    slider2 = FloatRangeSlider(min=min_dict[drop2.value],
                                               max=max_dict[drop2.value],
                                               value=[visminG, vismaxG],
                                               step=0.01)
                    slider3 = FloatRangeSlider(min=min_dict[drop3.value],
                                               max=max_dict[drop3.value],
                                               value=[visminB, vismaxB],
                                               step=0.01)
                    # set handlers
                    drop.observe(handle_dropdown(slider), names=['value'])
                    drop2.observe(handle_dropdown(slider2), names=['value'])
                    drop3.observe(handle_dropdown(slider3), names=['value'])

                    # widget for band selector + slider
                    band_slider = HBox([drop, slider])
                    band_slider2 = HBox([drop2, slider2])
                    band_slider3 = HBox([drop3, slider3])

                    return VBox([band_slider, band_slider2, band_slider3],
                                layout=Layout(width='700px'))

                # Create widget for 1 or 3 bands
                bands = RadioButtons(options=['1 band', '3 bands'],
                                     layout=Layout(width='80px'))

                # Create widget for band, min and max selection
                selection = slider_1band()

                # Apply button
                apply = Button(description='Apply', layout=Layout(width='100px'))

                # new row
                new_row = [bands, selection, apply]

                # update row of widgets
                def update_row_items(new_row):
                    self.items = [[selector, group1],
                                   new_row]

                # handler for radio button (1 band / 3 bands)
                def handle_radio_button(change):
                    new = change['new']
                    if new == '1 band':
                        # create widget
                        selection = slider_1band() # TODO
                        # update row of widgets
                        update_row_items([bands, selection, apply])
                    else:
                        red = slider_1band(name='red') # TODO
                        green = slider_1band(name='green')
                        blue = slider_1band(name='blue')
                        selection = VBox([red, green, blue])
                        # selection = slider_3bands()
                        update_row_items([bands, selection, apply])

                def handle_apply(button):
                    radio = self.items[1][0].value # radio button
                    vbox = self.items[1][1]
                    print('vbox', vbox)
                    if radio == '1 band':  # 1 band
                        hbox_band = vbox.children[0].children

                        band = hbox_band[0].value
                        min = hbox_band[1].value[0]
                        max = hbox_band[1].value[1]

                        map.addLayer(image, {'bands':[band], 'min':min, 'max':max},
                                     layer_name)
                    else:  # 3 bands
                        hbox_bandR = vbox.children[0].children
                        hbox_bandG = vbox.children[1].children
                        hbox_bandB = vbox.children[2].children

                        bandR = hbox_bandR[0].value
                        bandG = hbox_bandG[0].value
                        bandB = hbox_bandB[0].value

                        minR = hbox_bandR[1].value[0]
                        minG = hbox_bandG[1].value[0]
                        minB = hbox_bandB[1].value[0]

                        maxR = hbox_bandR[1].value[1]
                        maxG = hbox_bandG[1].value[1]
                        maxB = hbox_bandB[1].value[1]

                        map.addLayer(image, {'bands':[bandR, bandG, bandB],
                                             'min':[float(minR), float(minG), float(minB)],
                                             'max':[float(maxR), float(maxG), float(maxB)]},
                                     layer_name)

                bands.observe(handle_radio_button, names='value')
                update_row_items(new_row)
                apply.on_click(handle_apply)


# coding=utf-8
""" Charts from Google Earth Engine data. Inpired by this question
https://gis.stackexchange.com/questions/291823/ui-charts-for-indices-time-series-in-python-api-of-google-earth-engine
and https://youtu.be/FytuB8nFHPQ, but at the moment relaying on `pygal`
library because it's the easiest to integrate with ipywidgets

author: Rodrigo E. Principe
author email: fitoprincipe82@gmail.com
gitHub: https://github.com/gee-community/gee_tools
"""
import pygal
import base64
import ee
from . import tools
import pandas as pd


class Line(pygal.Line):
    def __init__(self, **kwargs):
        super(Line, self).__init__(**kwargs)
        self.data = None
        self.y_labels = dict()

    @staticmethod
    def from_pandas(dataframe, x=None, y=None, datetime=False, **kwargs):
        """ Creates a Line chart from a pandas dataFrame """
        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError('first argument must be a pandas DataFrame')

        line_chart = Line(**kwargs)

        if not x:
            labels = dataframe[dataframe.columns[0]]
        else:
            labels = dataframe[x].values.tolist()

        if not datetime:
            x_values = labels
        else:
            x_values = [tools.Date.millis2datetime(d) for d in labels]

        line_chart.x_labels = x_values

        if isinstance(y, list):
            for column in y:
                if column == x:
                    continue
                ydata = dataframe[column].values.tolist()
                ydata = [{'value':data} for data in ydata]
                # TODO: add values config
                # pygal.org/en/latest/documentation/configuration/value.html
                line_chart.add(column, ydata)
                line_chart.y_labels[column] = ydata
        else:
            ydata = dataframe[y].values.tolist()
            line_chart.add(y, ydata)
            line_chart.y_labels[y] = ydata

        line_chart.data = dataframe
        return line_chart

    def render_widget(self, width=None, height=None):
        from ipywidgets import HTML

        b64 = base64.b64encode(self.render())
        src = 'data:image/svg+xml;charset=utf-8;base64,'+b64

        if width and not height:
            html = '<embed src={} width={}></embed>'.format(src, width)
        elif height and not width:
            html = '<embed src={} height={}></embed>'.format(src, height)
        elif width and height:
            html = '<embed src={} height={} width={}></embed>'.format(src, height, width)
        else:
            html = '<embed src={}>'.format(src)

        return HTML(html)


class Image(object):
    """ Charts for Images """
    def __init__(self, source):
        self.source = source

    @staticmethod
    def data2pandas(data):
        ''' Convert data coming from tools.get_values to a pandas DataFrame'''
        # Indices
        # header
        allbands = [val.keys() for bands, val in data.items()]
        header = []
        for bandlist in allbands:
            for band in bandlist:
                if band not in header:
                    header.append(band)

        data_dict = {}
        indices = []
        for i, head in enumerate(header):
            band_data = []
            for iid, val in data.items():
                if i == 0:
                    indices.append(iid)
                band_data.append(val[head])
            data_dict[head] = band_data

        df = pd.DataFrame(data=data_dict, index=indices)

        return df

    @staticmethod
    def check_imageCollection(imageCollection):
        if not isinstance(imageCollection, ee.ImageCollection):
            msg = 'first parameter of Image.doySeries must be an ' \
                  'ImageCollection, found {}'
            raise ValueError(msg.format(type(imageCollection)))

    @staticmethod
    def series(imageCollection, region, reducer=ee.Reducer.mean(),
               scale=None, xProperty='system:time_start', bands=None,
               labels=None):
        Image.check_imageCollection(imageCollection)

        # scale
        if not scale:
            scale = 1

        # first image (for getting bands and properties)
        first = ee.Image(imageCollection.first())
        allbands = first.bandNames().getInfo()

        # Get Y (bands)
        if not bands:
            ydata = allbands
        else:
            ydata = bands

        # Get Images properties
        properties = first.propertyNames().getInfo()

        # If xProperty == 'system:time_start' will compute datetime
        datetime = True if xProperty == 'system:time_start' else False

        # generate data
        if isinstance(region, ee.Geometry):
            geom = region
        elif isinstance(region, (ee.Feature, ee.FeatureCollection)):
            geom = region.geometry()
        else:
            msg = 'Parameter `region` must be `ee.Geometry`, `ee.Feautre` or' \
                  ' or `ee.FeatureCollection, found {}'
            raise ValueError(msg.format(type(region)))

        if xProperty in properties:
            # include xProperty in data
            x_property = [xProperty]
        elif xProperty in allbands:
            # xProperty will be included in data because is a band
            x_property = []
        else:
            msg = 'xProperty "{}" not found in properties or bands'
            raise ValueError(msg.format(xProperty))

        data = tools.get_values(imageCollection, geom, reducer, scale,
                                properties=x_property, side='client')

        # Replace band names with labels provided
        if labels and len(ydata) == len(labels):
            for iid, values_dict in data.items():
                for old_name, new_name in zip(ydata, labels):
                    data[iid][new_name] = data[iid][old_name]
                    data[iid].pop(old_name)
            ydata = labels

        df = Image.data2pandas(data)
        newdf = df.sort_values(xProperty)

        line_chart = Line.from_pandas(newdf, y=ydata,
                                      x=xProperty, datetime=datetime)

        reducer_name = reducer.getInfo()['type'].split('.')[1]
        chart_title = 'Band {} across images'.format(reducer_name)
        line_chart.title = chart_title

        return line_chart

    @staticmethod
    def seriesByRegion(imageCollection, regions, reducer, band=None,
                       scale=None, xProperty='system:time_start',
                       seriesProperty='system:index'):

        # If xProperty == 'system:time_start' will compute datetime
        datetime = True if xProperty == 'system:time_start' else False

        Image.check_imageCollection(imageCollection)
        first = ee.Image(imageCollection.first())

        # Make defaults
        if not band:
            band = ee.String(first.bandNames().get(0))
        if not scale:
            # scale = first.select([0]).projection().nominalScale()
            scale = 1

        imageCollection = imageCollection.select([band])

        # Generate data
        # Geometry
        if isinstance(regions, ee.Geometry):
            print('Using `seriesByRegion` with `ee.Geometry` will give you'
                  ' the same output as `series`, use that method instead')
            chart_title = '{} values in merged geometry across images'.format(band)

            chart_line = Image.series(imageCollection, regions, reducer, scale=scale,
                                      xProperty=xProperty, bands=[band], labels=['geometry'])
            chart_line.title = chart_title
            return chart_line

        elif isinstance(regions, ee.Feature):
            reducer_name = reducer.getInfo()['type'].split('.')[1]
            chart_title = '{} {} values in one regions across images\nlabeled by {}'.format(
                band, reducer_name, seriesProperty)
            label = regions.get(seriesProperty).getInfo()
            label = label if label else 'unknown feature'
            chart_line = Image.series(imageCollection, regions, reducer, scale=scale,
                                      xProperty=xProperty, bands=[band], labels=[label])
            chart_line.title = chart_title
            return chart_line

        elif isinstance(regions, ee.FeatureCollection):

            def over_col(img, inicol):
                inicol = ee.Dictionary(inicol)
                x_prop = img.get(xProperty)
                iid = img.get('system:index')

                def over_fc(feat, inifeat):
                    inifeat = ee.Dictionary(inifeat)
                    name = feat.get(seriesProperty)
                    data = img.reduceRegion(reducer,
                                            feat.geometry(),
                                            scale).get(band)
                    return inifeat.set(name, data)

                fc_data = ee.Dictionary(
                    regions.iterate(over_fc, ee.Dictionary({}))
                ).set(xProperty, x_prop)

                return inicol.set(iid, fc_data)

            data = ee.Dictionary(
                imageCollection.iterate(over_col, ee.Dictionary({})))

            data = data.getInfo()

            df = Image.data2pandas(data)
            newdf = df.sort_values(xProperty)

            y_labels = newdf.columns.values.tolist()

            line_chart = Line.from_pandas(newdf, y=y_labels,
                                          x=xProperty, datetime=datetime)

            reducer_name = reducer.getInfo()['type'].split('.')[1]
            chart_title = '{} {} values in different regions across images\nlabeled by {}'.format(
                band, reducer_name, seriesProperty)
            line_chart.title = chart_title

            return line_chart




# coding=utf-8

''' This module is designed to use in the Jupyter Notebook or generating Maps
for Web applications. It uses folium and branca, and is inspired in
https://github.com/mccarthyryanc/folium_gee '''

from __future__ import print_function
import folium
from folium import features
import ee
from copy import copy
from . import tools
import json
import math
from uuid import uuid4

if not ee.data._initialized: ee.Initialize()


class Map(folium.Map):

    def __init__(self, **kwargs):
        super(Map, self).__init__(**kwargs)
        self.added_geometries = 0

    def show(self):
        LC = folium.LayerControl()
        self.add_child(LC)
        return self

    def addMarker(self, point, visParams=None, data=None, scale=None):
        """ Adds a marker with a popup given the data from an Image if
        data param is specified

        :param point: point
        :type point: ee.Geometry or ee.Feature
        :param visParams: visualization parameters
        :type visParams: dict
        :param data: Image from where to get the data from
        :type data: ee.Image
        :return:
        """
        if isinstance(point, ee.Geometry):
            geometry = point
        elif isinstance(point, ee.Feature):
            geometry = point.geometry()

        info = geometry.getInfo()
        type = info['type']
        coords = info['coordinates']
        coords = inverse_coordinates(coords)

        if data:
            # Try to get the image scale
            scale = data.projection().nominalScale().getInfo() if not scale else scale

            # Reduce if computed scale is too big
            scale = 1 if scale > 500 else scale

            values = tools.get_value(data, geometry, scale, 'client')
            val_str = ''
            for key, val in values.iteritems():
                val_str += '<b>{}:</b> {}</br>'.format(key, val)
                marker = folium.Marker(location=coords,
                                       popup=folium.Popup(val_str))
        else:
            marker = folium.Marker(location=coords)

        self.add_child(marker)

    def addLayer(self, eeObject, visParams=None, name=None, show=True,
                 opacity=None, inspect={'data':None, 'reducer':None, 'scale':None}):
        """
        Adds a given EE object to the map as a layer.

        Returns the new map layer.

        Arguments:
        eeObject (Collection|Feature|Image|MapId):
            The object to add to the map.

        visParams (FeatureVisualizationParameters|ImageVisualizationParameters,
        optional):
            The visualization parameters. For Images and ImageCollection,
            see ee.data.getMapId for valid parameters. For Features and
            FeatureCollections, the only supported key is "color",
            as a CSS 3.0 color string or a hex string in "RRGGBB" format.

        name (String, optional):
            The name of the layer. Defaults to "Layer N".

        shown (Boolean, optional):
            A flag indicating whether the layer should be on by default.

        opacity (Number, optional):
            The layer's opacity represented as a number between 0 and 1.
            Defaults to 1.

        Returns: ui.Map.Layer
        """

        def addImage(image):
            params = get_image_tile(image, visParams, show, opacity)
            layer = folium.TileLayer(attr=params['attribution'],
                                     name=name,
                                     overlay=params['overlay'],
                                     tiles=params['url'])
            layer.add_to(self)
            return self

        def addGeoJson(geometry):
            params = get_geojson_tile(geometry, name, inspect)
            geojson = json.dumps(params['geojson'])
            layer = features.GeoJson(geojson,
                                     name=name)
            pop = folium.Popup(params['pop'])

            layer.add_child(pop)

            self.added_geometries += 1
            self.add_child(layer)

        # CASE: ee.Image
        if isinstance(eeObject, ee.Image):
            addImage(eeObject)

        elif isinstance(eeObject, ee.Geometry):
            addGeoJson(eeObject)

        elif isinstance(eeObject, ee.Feature):
            geom = eeObject.geometry()
            addGeoJson(geom)

        elif isinstance(eeObject, ee.ImageCollection):
            pass
        else:
            raise ValueError('addLayer currently supports ee.Image as eeObject argument')

    def centerObject(self, eeObject, zoom=None):
        """
        Centers the map view on a given object.

        Returns the map.

        Arguments:
        object (Element|Geometry):
        An object to center on - a geometry, image or feature.

        zoom (Number, optional):
        The zoom level, from 1 to 24. If unspecified, computed based on the object's bounding box.
        :return:
        """
        bounds = get_bounds(eeObject)[0]
        self.fit_bounds([bounds[0], bounds[2]], max_zoom=zoom)

        return self

def get_bounds(eeObject):
    if isinstance(eeObject, list):
        bounds = eeObject
    else:
        # Make a buffer if object is a Point
        if isinstance(eeObject, ee.Geometry):
            t = eeObject.type().getInfo()
            if t == 'Point':
                eeObject = eeObject.buffer(1000)

        bounds = tools.getRegion(eeObject, True)

    # Catch unbounded images
    unbounded = [[[-180.0, -90.0], [180.0, -90.0],
                  [180.0, 90.0], [-180.0, 90.0],
                  [-180.0, -90.0]]]

    if bounds == unbounded:
        print("can't center object because it is unbounded")
        return None

    bounds = inverse_coordinates(bounds)
    return bounds

def get_default_vis(image, stretch=0.8):
    bandnames = image.bandNames().getInfo()

    if len(bandnames) < 3:
        selected = image.select([0]).getInfo()
        bandnames = bandnames[0]
    else:
        selected = image.select([0, 1, 2]).getInfo()
        bandnames = [bandnames[0], bandnames[1], bandnames[2]]

    bands = selected['bands']
    # bandnames = [bands[0]['id'], bands[1]['id'], bands[2]['id']]
    types = bands[0]['data_type']

    maxs = {'float':1,
            'double': 1,
            'int8': 127, 'uint8': 255,
            'int16': 32767, 'uint16': 65535,
            'int32': 2147483647, 'uint32': 4294967295,
            'int64': 9223372036854776000}

    precision = types['precision']

    if precision == 'float':
        btype = 'float'
    elif precision == 'double':
        btype = 'double'
    elif precision == 'int':
        max = types['max']
        maxs_inverse = dict((val, key) for key, val in maxs.iteritems())
        btype = maxs_inverse[int(max)]
    else:
        raise ValueError('Unknown data type {}'.format(precision))

    limits = {'float': 0.8}

    for key, val in maxs.iteritems():
        limits[key] = val*stretch

    min = 0
    max = limits[btype]
    return {'bands':bandnames, 'min':min, 'max':max}

def is_point(item):
    """ Determine if the given list has the structure of a point. This is:
    it is a list or tuple with two int or float items """
    if isinstance(item, list) or isinstance(item, tuple):
        if len(item) == 2:
            lon = item[0]
            if isinstance(lon, int) or isinstance(lon, float):
                return True
            else:
                return False
        else:
            return False
    else:
        return False

def inverse_coordinates(coords):
    """ Inverse a set of coordinates (any nesting depth)

    :param coords: a nested list of points
    :type coords: list
    """
    newlist = []
    if is_point(coords):
        return [coords[1], coords[0]]
    elif not isinstance(coords, list) and not isinstance(coords, tuple):
        raise ValueError('coordinates to inverse must be minimum a point')
    for i, it in enumerate(coords):
        p = is_point(it)
        if not p and (isinstance(it, list) or isinstance(it, tuple)):
            newlist.append(inverse_coordinates(it))
        else:
            newp = [it[1],it[0]]
            newlist.append(newp)
    return newlist

def visparams_str2list(params):
    ''' Transform a string formated as needed by ee.data.getMapId to a list

    :param params: params to convert
    :type params: str
    :return: a list with the params
    :rtype: list
    '''
    proxy_bands = []
    bands = params.split(',')
    for band in bands:
        proxy_bands.append(band.strip())
    return proxy_bands

def visparams_list2str(params):
    ''' Transform a list to a string formated as needed by
        ee.data.getMapId

    :param params: params to convert
    :type params: list
    :return: a string formated as needed by ee.data.getMapId
    :rtype: str
    '''
    n = len(params)
    if n == 1:
        newbands = '{}'.format(params[0])
    elif n == 3:
        newbands = '{},{},{}'.format(params[0], params[1], params[2])
    else:
        newbands = '{}'.format(params[0])
    return newbands

def get_image_tile(image, visParams, show=True, opacity=None,
                   overlay=True):

    proxy = {}
    params = visParams if visParams else {}

    # BANDS #############
    def default_bands(image):
        bandnames = image.bandNames().getInfo()
        if len(bandnames) < 3:
            bands = [bandnames[0]]
        else:
            bands = [bandnames[0], bandnames[1], bandnames[2]]
        return bands
    bands = params.get('bands') if 'bands' in params else default_bands(image)

    # if the passed bands is a string formatted like required by GEE, get the
    # list out of it
    if isinstance(bands, str):
        bands_list = visparams_str2list(bands)
        bands_str = visparams_list2str(bands_list)

    # Transform list to getMapId format
    # ['b1', 'b2', 'b3'] == 'b1, b2, b3'
    if isinstance(bands, list):
        bands_list = bands
        bands_str = visparams_list2str(bands)

    # Set proxy parameteres
    proxy['bands'] = bands_str

    # MIN #################
    themin = params.get('min') if 'min' in params else '0'

    # if the passed min is a list, convert to the format required by GEE
    if isinstance(themin, list):
        themin = visparams_list2str(themin)

    proxy['min'] = themin

    # MAX #################
    def default_max(image, bands):
        proxy_maxs = []
        maxs = {'float':1,
                'double': 1,
                'int8': ((2**8)-1)/2, 'uint8': (2**8)-1,
                'int16': ((2**16)-1)/2, 'uint16': (2**16)-1,
                'int32': ((2**32)-1)/2, 'uint32': (2**32)-1,
                'int64': ((2**64)-1)/2}
        for band in bands:
            ty = image.select([band]).getInfo()['bands'][0]['data_type']
            try:
                themax = maxs[ty]
            except:
                themax = 1
            proxy_maxs.append(themax)
        return proxy_maxs

    themax = params.get('max') if 'max' in params else default_max(image,
                                                                   bands_list)

    # if the passed max is a list or the max is computed by the default function
    # convert to the format required by GEE
    if isinstance(themax, list):
        themax = visparams_list2str(themax)

    proxy['max'] = themax

    # PALETTE ################
    if 'palette' in params:
        if len(bands_list) == 1:
            palette = params.get('palette')
            if isinstance(palette, str):
                palette = visparams_str2list(palette)
            toformat = '{},'*len(palette)
            palette = toformat[:-1].format(*palette)
            proxy['palette'] = palette
        else:
            print("Can't use palette parameter with more than one band")

    # Get the MapID and Token after applying parameters
    image_info = image.getMapId(proxy)
    mapid = image_info['mapid']
    token = image_info['token']
    tiles = "https://earthengine.googleapis.com/map/%s/{z}/{x}/{y}?token=%s"%(mapid,token)
    attribution = 'Map Data &copy; <a href="https://earthengine.google.com/">Google Earth Engine</a> '
    overlay = overlay

    return {'url': tiles,
            'attribution': attribution,
            'overlay': overlay,
            'show': show,
            'opacity': opacity,
            'visParams': proxy,
            }

def feature_properties_output(feat):
    ''' generates a string for features properties '''
    info = feat.getInfo()
    properties = info['properties']
    theid = info['id']
    stdout = '<h3>ID {}</h3></br>'.format(theid)
    for prop, value in properties.items():
        try:
            value = str(value)
        except:
            try:
                value = value.encode('utf-8', errors='ignore')
            except:
                value = value.encode('latin1', errors='ignore')

        try:
            prop = str(prop)
        except:
            try:
                prop = prop.encode('utf-8', errors='ignore')
            except:
                prop = prop.encode('latin1', errors='ignore')

        stdout += '<b>{}</b>: {}</br>'.format(prop, value)
    return stdout

def get_geojson_tile(geometry, name=None,
                     inspect={'data':None, 'reducer':None, 'scale':None}):
    ''' Get a GeoJson giving a ee.Geometry or ee.Feature '''

    if isinstance(geometry, ee.Feature):
        feat = geometry
        geometry = feat.geometry()
    else:
        feat = None

    info = geometry.getInfo()
    type = info['type']

    gjson_types = ['Polygon', 'LineString', 'MultiPolygon',
                   'LinearRing', 'MultiLineString', 'MultiPoint',
                   'Point', 'Polygon', 'Rectangle',
                   'GeometryCollection']

    # newname = name if name else "{} {}".format(type, map.added_geometries)

    if type in gjson_types:
        data = inspect['data']
        if feat:
            default_popup = feature_properties_output(feat)
        else:
            default_popup = type
        red = inspect.get('reducer','first')
        sca = inspect.get('scale', None)
        popval = get_data(geometry, data, red, sca, name) if data else default_popup
        geojson = geometry.getInfo()

        return {'geojson':geojson,
                'pop': popval}
                # 'name': newname}
    else:
        print('unrecognized object type to add to map')

def get_zoom(bounds, method=1):
    '''
    as ipyleaflet does not have a fit bounds method, try to get the zoom to fit

    from: https://stackoverflow.com/questions/6048975/google-maps-v3-how-to-calculate-the-zoom-level-for-a-given-bounds
    '''
    bounds = bounds[0]
    sw = bounds[0]
    ne = bounds[2]

    sw_lon = sw[1]
    sw_lat = sw[0]
    ne_lon = ne[1]
    ne_lat = ne[0]

    def method1():
        # Method 1
        WORLD_DIM = {'height': 256, 'width': 256}
        ZOOM_MAX = 21

        def latRad(lat):
            sin = math.sin(lat * math.pi / 180)
            radX2 = math.log((1 + sin) / (1 - sin)) / 2
            return max(min(radX2, math.pi), -math.pi) / 2

        def zoom(mapPx, worldPx, fraction):
            return math.floor(math.log(mapPx / worldPx / fraction) / math.log(2))

        latFraction = float(latRad(ne_lat) - latRad(sw_lat)) / math.pi

        lngDiff = ne_lon - sw_lon

        lngFraction = (lngDiff + 360) if (lngDiff < 0) else lngDiff
        lngFraction = lngFraction / 360

        latZoom = zoom(400, WORLD_DIM['height'], latFraction)
        lngZoom = zoom(970, WORLD_DIM['width'], lngFraction)

        return int(min(latZoom, lngZoom, ZOOM_MAX))

    def method2():
        scale = 111319.49

        GLOBE_WIDTH = 256 # a constant in Google's map projection

        angle = ne_lon - sw_lon
        if angle < 0:
            angle += 360
        zoom = math.floor(math.log(scale * 360 / angle / GLOBE_WIDTH) / math.log(2))
        return int(zoom)-8

    finalzoom = method1() if method == 1 else method2()
    return finalzoom

# TODO: Multiple dispatch! https://www.artima.com/weblogs/viewpost.jsp?thread=101605
def get_data(geometry, obj, reducer='first', scale=None, name=None):
    ''' Get data from an ee.ComputedObject using a giving ee.Geometry '''
    accepted = (ee.Image, ee.ImageCollection, ee.Feature, ee.FeatureCollection)

    reducers = {'first': ee.Reducer.first(),
                'mean': ee.Reducer.mean(),
                'median': ee.Reducer.median(),
                'sum':ee.Reducer.sum()}

    if not isinstance(obj, accepted):
        return "Can't get data from that Object"
    elif isinstance(obj, ee.Image):
        t = geometry.type().getInfo()
        # Try to get the image scale
        scale = obj.select([0]).projection().nominalScale().getInfo()\
                if not scale else scale

        # Reduce if computed scale is too big
        scale = 1 if scale > 500 else scale
        if t == 'Point':
            values = tools.get_value(obj, geometry, scale, 'client')
            val_str = '<h3>Data from {}'.format(name)
            for key, val in values.iteritems():
                val_str += '<b>{}:</b> {}</br>'.format(key, val)
            return val_str
        elif t == 'Polygon':
            red = reducer if reducer in reducers.keys() else 'first'
            values = obj.reduceRegion(reducers[red], geometry, scale, maxPixels=1e13).getInfo()
            val_str = '<h3>{}:</h3>\n'.format(red)
            for key, val in values.iteritems():
                val_str += '<b>{}:</b> {}</br>'.format(key, val)
            return val_str

def create_html(dictionary, nest=0, ini='', indent=2):
    """ Create a HTML output from a dict object

    :param nest: nesting level to start (defaults to 0)
    :type nest: int
    :param ini: initial text
    :type ini: str
    :param indent: indentation spaces
    :type indent: int
    :return: the text to use in a HTML object
    :rtype: str
    """

    for key, val in dictionary.items():
        if isinstance(val, dict):
            line = '{}<b>{}</b>:</br>'.format('&nbsp;'*nest, key)
            ini += line
            newnest = nest+indent
            ini = create_html(val, newnest, ini)
        elif isinstance(val, list):
            # tranform list to a dictionary
            dictval = {k: v for k, v in enumerate(val)}
            line = '{}<b>{}</b>:</br>'.format('&nbsp;'*nest, key)
            ini += line
            newnest = nest+indent
            ini = create_html(dictval, newnest, ini)
        else:
            line = '{}<b>{}</b>: {}</br>'.format('&nbsp;'*nest, key, val)
            ini += line

    return ini

def create_html_table(header, rows):
    ''' Create a HTML table

    :param header: a list of headers
    :type header: list
    :param rows: a list with lists of values
    :type rows: list of lists
    :return: a HTML string
    :rtype: str
    '''
    uid = 'a'+uuid4().hex
    style = '<style>\n.{}\n{{border: 1px solid black;\n border-collapse: collapse;}}\n</style>\n'
    style = style.format(uid)
    general = '<table class="{}">\n{{}}</table>'.format(uid)
    row = '  <tr class="{}">\n{{}}</tr>\n'.format(uid)
    col = '    <td style="text-align: center" class="{}">{{}}</td>\n'.format(uid)
    headcol = '    <th class="{}">{{}}</th>\n'.format(uid)

    # header
    def create_row(alist, template):
        cols = ''
        for el in alist:
            cols += template.format(el)
        newrow = row.format(cols)
        return newrow

    header_row = create_row(header, headcol)

    rest = ''
    # rows
    for r in rows:
        newrow = create_row(r, col)
        rest += newrow

    body = '{}\n{}'.format(header_row, rest)
    html = '{}\n{}'.format(style, general.format(body))

    return html


def paint(geometry, outline_color='black', fill_color=None, outline=2):
    ''' Paint a Geometry, Feature or FeatureCollection '''

    def overlap(image_back, image_front):
        mask_back = image_back.mask()
        mask_front = image_front.mask()
        entire_mask = mask_back.add(mask_front)
        mask = mask_back.Not()
        masked = image_front.updateMask(mask).unmask()

        return masked.add(image_back.unmask()).updateMask(entire_mask)

    if isinstance(geometry, ee.Feature) or isinstance(geometry, ee.FeatureCollection):
        geometry = geometry.geometry()

    if fill_color:
        fill = ee.Image().paint(geometry, 1).visualize(palette=[fill_color])

    if outline_color:
        out = ee.Image().paint(geometry, 1, outline).visualize(palette=[outline_color])

    if fill_color and outline_color:
        rgbVector = overlap(out, fill)
    elif fill_color:
        rgbVector = fill
    else:
        rgbVector = out

    return rgbVector



# coding=utf-8
import unittest
import ee
from .. import tools, cloud_mask
ee.Initialize()

area = ee.Geometry.Point([-65.80, -25.01])

l1 = ee.Image("LANDSAT/LM1_L1T/LM12500851973064AAA05")
l2 = ee.Image("LANDSAT/LM2_L1T/LM22480891976038FAK05")
l3 = ee.Image("LANDSAT/LM3_L1T/LM32490891979068XXX01")
l5led = ee.Image("LEDAPS/LT5_L1T_SR/LT52310901984169XXX03")
l7led = ee.Image("LEDAPS/LE7_L1T_SR/LE72300781999227CUB00")
sentinel2 = ee.Image("COPERNICUS/S2/20160122T144426_20160127T174646_T18GYT")
modis_terra = ee.Image("MODIS/MOD09GA/MOD09GA_005_2000_02_24")
modis_aqua = ee.Image("MODIS/MYD09GA/MYD09GA_005_2002_07_04")

p_l8SR_cloud = ee.Geometry.Point([-65.8109, -25.0185])
p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

bands457 = ['B4', 'B7', 'B3']
bands8 = ['B5', 'B7', 'B4']
bandsS2 = ['B8', 'B12', 'B4']

def getimage(collection, date, area):
    image = ee.Image(collection.filterDate(ee.Date(date).advance(-1, 'day'),
                                           ee.Date(date).advance(1, 'day'))
                     .filterBounds(area)
                     .first())
    return image

def show(iid, image, masked, bands, min, max):
    thumb_original = image.select(bands).getThumbUrl(
        {'min':min, 'max':max, 'region':tools.getRegion(image)})
    thumb_masked = masked.select(bands).getThumbUrl(
        {'min':min, 'max':max, 'region':tools.getRegion(image)})

    print('{}\noriginal: {}\nmasked {}\n'.format(iid, thumb_original, thumb_masked))


class TestL4TOA(unittest.TestCase):
    def setUp(self):
        self.iid = 'LANDSAT/LT04/C01/T1_TOA'
        collection = ee.ImageCollection(self.iid)
        date = '1989-07-25'
        self.p_cloud = ee.Geometry.Point([-65.506, -24.9263])
        p_clear = ee.Geometry.Point([-64.6133, -24.4522])

        self.image = getimage(collection, date, area)

        print '{}/{}'.format(self.iid, self.image.id().getInfo())

    def test_all(self):
        masked = cloud_mask.landsatTOA()(self.image)
        vals = tools.get_value(masked, self.p_cloud, 30, 'client')

        show(self.iid, self.image, masked, bands457, 0, 0.5)

        self.assertEqual(vals["B1"], None)

    def test_clouds(self):
        masked = cloud_mask.landsatTOA(['cloud'])(self.image)
        vals = tools.get_value(masked, self.p_cloud, 30, 'client')

        show(self.iid, self.image, masked, bands457, 0, 0.5)

        self.assertEqual(vals["B1"], None)

    def test_shadows(self):
        masked = cloud_mask.landsatTOA(['shadow'])(self.image)
        vals = tools.get_value(masked, self.p_cloud, 30, 'client')

        show(self.iid, self.image, masked, bands457, 0, 0.5)

        # self.assertEqual(vals["B1"], None)

    def test_snow(self):
        masked = cloud_mask.landsatTOA(['snow'])(self.image)
        vals = tools.get_value(masked, self.p_cloud, 30, 'client')

        show(self.iid, self.image, masked, bands457, 0, 0.5)

        # self.assertEqual(vals["B1"], None)


class TestL5TOA(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LT05/C01/T1_TOA'
        collection = ee.ImageCollection(iid)
        date = '2010-03-21'
        p_cloud = ee.Geometry.Point([-65.2835, -24.3422])
        p_clear = ee.Geometry.Point([-64.6106, -24.7469])

        image = getimage(collection, date, area)

        masked = cloud_mask.landsatTOA()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bands457, 0, 0.5)

        self.assertEqual(vals["B1"], None)


class TestL7TOA(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LE07/C01/T1_TOA'
        collection = ee.ImageCollection(iid)
        date = '2010-03-13'
        p_cloud = ee.Geometry.Point([-65.5784, -24.773])
        p_clear = ee.Geometry.Point([-65.599, -24.9313])

        image = getimage(collection, date, area)

        masked = cloud_mask.landsatTOA()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bands457, 0, 0.5)

        self.assertEqual(vals["B1"], None)


class TestL8TOA(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LC08/C01/T1_TOA'
        collection = ee.ImageCollection(iid)
        date = '2017-03-08'
        p_cloud = ee.Geometry.Point([-65.7539, -25.0327])
        p_clear = ee.Geometry.Point([-66.034, -24.9512])

        image = getimage(collection, date, area)

        masked = cloud_mask.landsatTOA()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bands457, 0, 0.5)

        self.assertEqual(vals["B1"], None)


class TestL4SR(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LT04/C01/T1_SR'
        collection = ee.ImageCollection(iid)
        date = '1989-07-25'
        p_cloud = ee.Geometry.Point([-65.506, -24.9263])
        p_clear = ee.Geometry.Point([-64.6133, -24.4522])

        image = getimage(collection, date, area)

        # masked = cloud_mask.cfmask_bits(image)
        masked = cloud_mask.landsatSR()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bands457, 0, 5000)

        self.assertEqual(vals["B1"], None)


class TestL5SR(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LT05/C01/T1_SR'
        collection = ee.ImageCollection(iid)
        date = '2010-03-21'
        p_cloud = ee.Geometry.Point([-65.2835, -24.3422])
        p_clear = ee.Geometry.Point([-64.6106, -24.7469])

        image = getimage(collection, date, area)

        # masked = cloud_mask.cfmask_bits(image)
        masked = cloud_mask.landsatSR()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')
        show(iid, image, masked, bands457, 0, 5000)

        self.assertEqual(vals["B1"], None)


class TestL7SR(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LE07/C01/T1_SR'
        collection = ee.ImageCollection(iid)
        date = '2010-03-13'
        p_cloud = ee.Geometry.Point([-65.5784, -24.773])
        p_clear = ee.Geometry.Point([-65.599, -24.9313])

        image = getimage(collection, date, area)

        # masked = cloud_mask.cfmask_bits(image)
        masked = cloud_mask.landsatSR()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')
        show(iid, image, masked, bands457, 0, 5000)

        self.assertEqual(vals["B1"], None)


class TestL8SR(unittest.TestCase):
    def test(self):
        iid = 'LANDSAT/LC08/C01/T1_SR'
        collection = ee.ImageCollection(iid)
        date = '2017-03-08'
        p_cloud = ee.Geometry.Point([-65.7539, -25.0327])
        p_clear = ee.Geometry.Point([-66.034, -24.9512])

        image = getimage(collection, date, area)

        # masked = cloud_mask.cfmask_bits(image)
        masked = cloud_mask.landsatSR()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')
        show(iid, image, masked, bands8, 0, 5000)

        self.assertEqual(vals["B1"], None)

class TestSentinel2(unittest.TestCase):
    def test(self):
        iid = 'COPERNICUS/S2'
        collection = ee.ImageCollection(iid)
        date = '2017-03-07'
        p_cloud = ee.Geometry.Point([-65.84304, -24.82382])
        p_clear = ee.Geometry.Point([-65.88415, -24.82608])

        image = getimage(collection, date, area)

        masked = cloud_mask.sentinel2()(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bandsS2, 0, 5000)

        self.assertEqual(vals["B1"], None)

class TestHollstein(unittest.TestCase):
    def test(self):
        iid = 'COPERNICUS/S2'
        collection = ee.ImageCollection(iid)
        date = '2017-03-07'
        p_cloud = ee.Geometry.Point([-65.84304, -24.82382])
        p_clear = ee.Geometry.Point([-65.88415, -24.82608])

        image = getimage(collection, date, area)

        masked = cloud_mask.hollstein_S2(['cloud','snow','shadow', 'cirrus'])(image)
        vals = tools.get_value(masked, p_cloud, 30, 'client')

        show(iid, image, masked, bandsS2, 0, 5000)

        self.assertEqual(vals["B1"], None)


# !/usr/bin/env python
# coding=utf-8

from __future__ import print_function
from . import tools
from . import decision_tree
import ee

import ee.data
if not ee.data._initialized: ee.Initialize()

def compute(image, mask_band, bits, options=None, name_all='all_masks'):
    """ Compute bits using a specified band, a bit's relation and a list of
    options

    :param image: the image that holds the bit mask band
    :type image: ee.Image
    :param mask_band: the name of the band that holds the bits mask
    :type mask_band: str
    :param bits: relation between name and bit
    :type bits: dict
    :param options: list of 'bits' to compute. Example: ['cloud', 'snow']. If
        None, will use all keys of the relation's dict
    :type options: list
    :param name_all: name for the band that holds the final mask. Default:
        'all_masks'
    :type name_all: str
    :return: The computed mask
    :rtype: ee.Image
    """
    # cast params in case they are not EE objects
    bits_dict = ee.Dictionary(bits)
    opt = ee.List(options) if options else bits_dict.keys()
    image = ee.Image(image).select(mask_band)

    first = ee.Image.constant(0).select([0], [name_all]) # init image

    # function for iterate over the options
    def for_iterate(option, ini):
        i = ee.Image(ini) # cast ini
        all = i.select([name_all])

        # bits relation dict contains the option?
        cond = bits_dict.contains(option)

        def for_true():
            ''' function to execute if condition == True '''
            # get the mask for the option
            mask = tools.compute_bits(bits_dict.get(option),
                                      bits_dict.get(option),
                                      option)(image)

            # name the mask
            # mask = ee.Image(mask).select([0], [option])
            newmask = all.Or(mask)

            # return ee.Image(all.Or(mask)).addBands(mask)
            return tools.replace(i, name_all, newmask).addBands(mask)

        return ee.Image(ee.Algorithms.If(cond, for_true(), i))

    good_pix = ee.Image(opt.iterate(for_iterate, first))

    # return good_pix.Not()
    return good_pix

### DEPRECATED FUNCTIONS ###
# GENERIC APPLICATION OF MASKS
def apply_masks(masks):
    """
    Get a single mask from many

    :param masks: list of ee.Image
    :type masks: list | ee.List
    :return: the resulting mask
    :rtype: ee.Image
    """
    masks = ee.List(masks) if isinstance(masks, list) else masks
    first = ee.Image.constant(0)

    def compute(mask, first):
        first = ee.Image(first)
        return first.Or(mask)

    bad_pixels = ee.Image(masks.iterate(compute, first))
    good_pixels = bad_pixels.Not()

    return good_pixels

# MODIS
def modis_(img):
    """ DEPRECATED

    Function to use in MODIS Collection

    Use:

    `masked = collection.map(cloud_mask.modis)`
    """
    cmask = img.select("state_1km")
    cloud = tools.compute_bits_client(cmask, 0, 0, "cloud")
    mix = tools.compute_bits_client(cmask, 1, 1, "mix")
    shadow = tools.compute_bits_client(cmask, 2, 2, "shadow")
    cloud2 = tools.compute_bits_client(cmask, 10, 10, "cloud2")
    snow = tools.compute_bits_client(cmask, 12, 12, "snow")

    mask = (cloud
            .Or(mix)
            # .Or(shadow)  # Cloud shadow seems to be miscomputed (MODIS/MYD09GA/MYD09GA_005_2015_09_18)
            .Or(cloud2)
            .Or(snow)
            )

    return img.updateMask(mask.Not())

# SENTINEL 2
def sentinel2_(image):
    """ DEPRECATED

    Function to use in SENTINEL2 Collection

    Use:
    `masked = collection.map(cloud_mask.sentinel2)`
    """
    nubes = image.select("QA60")
    opaque = tools.compute_bits_client(nubes, 10, 10, "opaque")
    cirrus = tools.compute_bits_client(nubes, 11, 11, "cirrus")
    mask = opaque.Or(cirrus)
    result = image.updateMask(mask.Not())
    return result

# FMASK
def fmask(bandname="fmask"):
    """ DEPRECATED

    Function to use in Collections that have a quality band computed with
    fmask algorithm

    The use of this function is a little different from the other because it
    is a function that returns a function, so you must call it to return the
    function to use in map:

    `masked = collection.map(cloud_mask.fmask())`

    :param bandname: name of the band that holds the fmask information
    :type bandname: str
    :return: a function to use in map
    :rtype: function
    """

    def fmask(image):
        imgFmask = image.select(bandname)
        shadow = imgFmask.eq(3)
        snow = imgFmask.eq(4)
        cloud = imgFmask.eq(5)

        mask = shadow.Or(snow).Or(cloud)

        imgMask = image.updateMask(mask.Not())
        return imgMask
    return fmask

def usgs(image):
    """ DEPRECATED

    Function to use in Surface Reflectance Collections computed by USGS

    Use:

    `masked = collection.map(cloud_mask.usgs)`
    """
    image = fmask("cfmask")(image)
    cloud = image.select("sr_cloud_qa").neq(255)
    shad = image.select("sr_cloud_shadow_qa").neq(255)
    return image.updateMask(cloud).updateMask(shad)

def cfmask_bits(image):
    """ DEPRECATED

    Function to use in Landsat Surface Reflectance Collections:
    LANDSAT/LT04/C01/T1_SR, LANDSAT/LT05/C01/T1_SR, LANDSAT/LE07/C01/T1_SR,
    LANDSAT/LC08/C01/T1_SR

    Use:

    `masked = collection.map(cloud_mask.cfmask_bits)`
    """
    bands = image.bandNames()
    contains_sr = bands.contains('sr_cloud_qa')

    def sr():
        mask = image.select('sr_cloud_qa')
        cloud_mask = tools.compute_bits(mask, 1, 1, 'cloud')
        shadow_mask = tools.compute_bits(mask, 2, 2, 'shadow')
        adjacent_mask = tools.compute_bits(mask, 3, 3, 'adjacent')
        snow_mask = tools.compute_bits(mask, 4, 4, 'snow')

        good_pix = cloud_mask.eq(0).And(shadow_mask.eq(0)).And(snow_mask.eq(0)).And(adjacent_mask.eq(0))
        return good_pix

    def pix():
        mask = image.select('pixel_qa')
        cloud_mask = tools.compute_bits(mask, 5, 5, 'cloud')
        shadow_mask = tools.compute_bits(mask, 3, 3, 'shadow')
        snow_mask = tools.compute_bits(mask, 4, 4, 'snow')

        good_pix = cloud_mask.eq(0).And(shadow_mask.eq(0)).And(snow_mask.eq(0))
        return good_pix

    good_pix = ee.Algorithms.If(contains_sr, sr(), pix())

    result = image.updateMask(good_pix)

    return result

def landsatTOA_(masks=['cloud', 'shadow', 'snow']):
    ''' DEPRECATED

    Function to mask out clouds, shadows and snow in Landsat 4 5 7 8 TOA:
    LANDSAT/LT04/C01/T1_TOA, LANDSAT/LT05/C01/T1_TOA, LANDSAT/LE07/C01/T1_TOA
    and LANDSAT/LC08/C01/T1_TOA

    :param masks: list of mask to compute
    :type masks: list
    :return: the funtion to apply in a map algorithm
    :rtype: function
    '''
    options = ee.List(masks)

    def wrap(img):
        mask = img.select('BQA')
        cloud_mask = tools.compute_bits_client(mask, 4, 4, 'cloud')
        shadow_mask = tools.compute_bits_client(mask, 8, 8, 'shadow')
        snow_mask = tools.compute_bits_client(mask, 10, 10, 'snow')

        relation = ee.Dictionary({
            'cloud': cloud_mask,
            'shadow': shadow_mask,
            'snow': snow_mask
        })

        masks_list = tools.get_from_dict(options, relation)  # make a list of masks
        good_pix = apply_masks(masks_list)

        return img.updateMask(good_pix)

    return wrap
#############################

def modis(options=['cloud', 'mix', 'shadow', 'cloud2', 'snow'], name='modis_mask',
          addBands=False, updateMask=True):
    bits = {
        'cloud': 0,
        'mix': 1,
        'shadow': 2,
        'cloud2':10,
        'snow':12}

    mask_band = 'state_1km'

    options = ee.List(options)
    def wrap(image):
        good_pix = compute(image, mask_band, bits, options, name_all=name)

        mask = good_pix.select([name]).Not()

        if addBands and updateMask:
            return image.updateMask(mask).addBands(good_pix)
        elif addBands:
            return image.addBands(good_pix)
        elif updateMask:
            return image.updateMask(mask)
        else:
            return image

    return wrap

def sentinel2(options=['opaque', 'cirrus'], name='esa_mask',
              addBands=False, updateMask=True):
    """ ESA Cloud cover assessment

    :param options: category to mask out. Can be: 'opaque' or/and 'cirrus'
    :type options: list
    :return: the function to mask out clouds
    """

    rel = {'opaque': 10, 'cirrus':11}
    band = 'QA60'

    def wrap(img):
        good_pix = compute(img, band, rel, options, name_all=name)
        mask = good_pix.select([name]).Not()

        if addBands and updateMask:
            return img.updateMask(mask).addBands(good_pix)
        elif addBands:
            return img.addBands(good_pix)
        elif updateMask:
            return img.updateMask(mask)
        else:
            return img

    return wrap

# LEDAPS
def ledaps(image):
    """ Function to use in Surface Reflectance Collections computed by
    LEDAPS

    Use:

    `masked = collection.map(cloud_mask.ledaps)`
    """
    cmask = image.select('QA')

    valid_data_mask = tools.compute_bits(cmask, 1, 1, 'valid_data')
    cloud_mask = tools.compute_bits(cmask, 2, 2, 'cloud')
    snow_mask = tools.compute_bits(cmask, 4, 4, 'snow')

    good_pix = cloud_mask.eq(0).And(valid_data_mask.eq(0)).And(snow_mask.eq(0))
    result = image.updateMask(good_pix)

    return result

def landsatSR(options=['cloud', 'shadow', 'adjacent', 'snow'], name='sr_mask',
              addBands=False, updateMask=True):
    """ Function to use in Landsat Surface Reflectance Collections:
    LANDSAT/LT04/C01/T1_SR, LANDSAT/LT05/C01/T1_SR, LANDSAT/LE07/C01/T1_SR,
    LANDSAT/LC08/C01/T1_SR

    :param options: masks to apply. Options: 'cloud', 'shadow', 'adjacent',
        'snow'
    :type options: list
    :param name: name of the band that will hold the final mask. Default: 'toa_mask'
    :type name: str
    :param addBands: add all bands to the image. Default: False
    :type addBands: bool
    :param updateMask: update the mask of the Image. Default: True
    :type updateMask: bool
    :return: a function for applying the mask
    :rtype: function
    """
    sr = {'bits': ee.Dictionary({'cloud': 1, 'shadow': 2, 'adjacent': 3, 'snow': 4}),
          'band': 'sr_cloud_qa'}

    pix = {'bits': ee.Dictionary({'cloud': 5, 'shadow': 3, 'snow': 4}),
           'band': 'pixel_qa'}

    # Parameters
    options = ee.List(options)

    def wrap(image):
        bands = image.bandNames()
        contains_sr = bands.contains('sr_cloud_qa')
        good_pix = ee.Image(ee.Algorithms.If(contains_sr,
                   compute(image, sr['band'], sr['bits'], options, name_all=name),
                   compute(image, pix['band'], pix['bits'], options, name_all=name)))

        mask = good_pix.select([name]).Not()

        if addBands and updateMask:
            return image.updateMask(mask).addBands(good_pix)
        elif addBands:
            return image.addBands(good_pix)
        elif updateMask:
            return image.updateMask(mask)
        else:
            return image

    return wrap

def landsatTOA(options=['cloud', 'shadow', 'snow'], name='toa_mask',
               addBands=False, updateMask=True):
    """ Function to mask out clouds, shadows and snow in Landsat 4 5 7 8 TOA:
    LANDSAT/LT04/C01/T1_TOA, LANDSAT/LT05/C01/T1_TOA, LANDSAT/LE07/C01/T1_TOA
    and LANDSAT/LC08/C01/T1_TOA

    :param options: masks to apply. Options: 'cloud', 'shadow', 'snow'
    :type options: list
    :param name: name of the band that will hold the final mask. Default: 'toa_mask'
    :type name: str
    :param addBands: add all bands to the image. Default: False
    :type addBands: bool
    :param updateMask: update the mask of the Image. Default: True
    :type updateMask: bool
    :return: a function for applying the mask
    :rtype: function
    """
    bits = ee.Dictionary({'cloud': 4, 'shadow': 8, 'snow': 10})
    mask_band = 'BQA'

    # Parameters
    opt = options if options else bits.keys()
    options = ee.List(opt)

    def wrap(image):
        good_pix = compute(image, mask_band, bits, options, name_all=name)

        mask = good_pix.select([name]).Not()

        if addBands and updateMask:
            return image.updateMask(mask).addBands(good_pix)
        elif addBands:
            return image.addBands(good_pix)
        elif updateMask:
            return image.updateMask(mask)
        else:
            return image

    return wrap

def hollstein_S2(options=['cloud', 'snow', 'shadow', 'water', 'cirrus'],
                 name='hollstein', addBands=False, updateMask=True):
    """

    :param options: masks to apply. Options: 'cloud', 'shadow', 'snow',
        'cirrus', 'water'
    :type options: list
    :param name: name of the band that will hold the final mask. Default: 'hollstein'
    :type name: str
    :param addBands: add all bands to the image. Default: False
    :type addBands: bool
    :param updateMask: update the mask of the Image. Default: True
    :type updateMask: bool
    :return: a function for applying the mask
    :rtype: function
    """

    def difference(a, b):
        def wrap(img):
            return img.select(a).subtract(img.select(b))
        return wrap

    def ratio(a, b):
        def wrap(img):
            return img.select(a).divide(img.select(b))
        return wrap

    def compute_dt(img):

        # 1
        b3 = img.select('B3').lt(3190)

        # 2
        b8a = img.select('B8A').lt(1660)
        r511 = ratio('B5', 'B11')(img).lt(4.33)

        # 3
        s1110 = difference('B11', 'B10')(img).lt(2550)
        b3_3 = img.select('B3').lt(5250)
        r210 = ratio('B2','B10')(img).lt(14.689)
        s37 = difference('B3', 'B7')(img).lt(270)

        # 4
        r15 = ratio('B1', 'B5')(img).lt(1.184)
        s67 = difference('B6', 'B7')(img).lt(-160)
        b1 = img.select('B1').lt(3000)
        r29 =  ratio('B2', 'B9')(img).lt(0.788)
        s911 = difference('B9', 'B11')(img).lt(210)
        s911_2 = difference('B9', 'B11')(img).lt(-970)

        snow = {'snow':[['1',0], ['22',0], ['34',0]]}
        cloud = {'cloud-1':[['1',0], ['22',1],['33',1],['44',1]],
                 'cloud-2':[['1',0], ['22',1],['33',0],['45',0]]}
        cirrus = {'cirrus-1':[['1',0], ['22',1],['33',1],['44',0]],
                  'cirrus-2':[['1',1], ['21',0],['32',1],['43',0]]}
        shadow = {'shadow-1':[['1',1], ['21',1],['31',1],['41',0]],
                  'shadow-2':[['1',1], ['21',1],['31',0],['42',0]],
                  'shadow-3':[['1',0], ['22',0],['34',1],['46',0]]}
        water = {'water':[['1',1], ['21',1],['31',0],['42',1]]}

        all = {'cloud':cloud,
               'snow': snow,
               'shadow':shadow,
               'water':water,
               'cirrus':cirrus}

        final = {}

        for option in options:
            final.update(all[option])

        dtf = decision_tree.binary(
                        {'1':b3,
                         '21':b8a, '22':r511,
                         '31':s37, '32':r210, '33':s1110, '34':b3_3,
                         '41': s911_2, '42':s911, '43':r29, '44':s67, '45':b1, '46':r15
                         }, final, name)

        results = dtf

        if updateMask and addBands:
            return img.addBands(results).updateMask(results.select(name))
        elif addBands:
            return img.addBands(results)
        elif updateMask:
            return img.updateMask(results.select(name))

    return compute_dt


# coding=utf-8

from __future__ import absolute_import, division, print_function

__version__ = "0.0.16"

__title__ = "geetools"
__summary__ = "A set of useful tools to use with Google Earth Engine Python" \
              "API"
__uri__ = "http://geetools.readthedocs.io"

__author__ = "Rodrigo E. Principe"
__email__ = "fitoprincipe82@gmail.com"

__license__ = "MIT"
__copyright__ = "2017 Rodrigo E. Principe"


# coding=utf-8
import unittest
import ee
from .. import imagestrip
from PIL.Image import Image
ee.Initialize()

class TestStrip(unittest.TestCase):

    def setUp(self):
        self.pol_L8SR = ee.Geometry.Polygon([[[-66, -25],
                                              [-66, -24.5],
                                              [-65.5, -24.5],
                                              [-65.5, -25]]])

        self.l8SR_col = ee.ImageCollection("LANDSAT/LC8_SR")

    def test_strip(self):
        strip = imagestrip.ImageStrip("test", description="just a test")

        col = self.l8SR_col.filterBounds(self.pol_L8SR).filterDate(
            "2013-01-01", "2013-06-01")

        list_imgs = col.toList(10)

        viz_params = {'bands':["B4", "B5", "B3"], 'min':0, 'max':5000}
        region = self.pol_L8SR.bounds().getInfo()["coordinates"]

        i = strip.from_collection([col], viz_param=viz_params, region=region,
                                  name="test", folder="files", drawRegion=True,
                                  zoom=2, properties=["CLOUD_COVER",
                                  "solar_zenith_angle"], description="test")

        self.assertIsInstance(i, Image)



# coding=utf-8

import unittest
import ee
# from geetools import tools
from .. import tools
ee.Initialize()

class TestTools(unittest.TestCase):

    def setUp(self):
        self.l1 = ee.Image("LANDSAT/LM1_L1T/LM12500851973064AAA05")
        self.l2 = ee.Image("LANDSAT/LM2_L1T/LM22480891976038FAK05")
        self.l3 = ee.Image("LANDSAT/LM3_L1T/LM32490891979068XXX01")
        self.l4toa = ee.Image("LANDSAT/LT4_L1T_TOA_FMASK/LT42310911992007XXX02")
        self.l5SR = ee.Image("LANDSAT/LT5_SR/LT52310772000341")
        self.l5led = ee.Image("LEDAPS/LT5_L1T_SR/LT52310901984169XXX03")
        self.l5toa = ee.Image("LANDSAT/LT5_L1T_TOA_FMASK/LT52320881999345COA00")
        self.l7SR = ee.Image("LANDSAT/LE7_SR/LE72300782012135")
        self.l7led = ee.Image("LEDAPS/LE7_L1T_SR/LE72300781999227CUB00")
        self.l7toa = ee.Image("LANDSAT/LE7_L1T_TOA_FMASK/LE72300781999227CUB00")
        self.l8toa = ee.Image("LANDSAT/LC8_L1T_TOA_FMASK/LC82310902013344LGN00")
        self.l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")
        self.sentinel2 = ee.Image("COPERNICUS/S2/20160122T144426_20160127T174646_T18GYT")
        self.modis_terra = ee.Image("MODIS/MOD09GA/MOD09GA_005_2000_02_24")
        self.modis_aqua = ee.Image("MODIS/MYD09GA/MYD09GA_005_2002_07_04")

        self.p_l8SR_cloud = ee.Geometry.Point([-65.8109, -25.0185])
        self.p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

        self.list1 = ee.List([1,2,3,4,5])
        self.list2 = ee.List([4,5,6,7])

        self.pol_L8SR = ee.Geometry.Polygon([[[-66, -25],
                                              [-66, -24.5],
                                              [-65.5, -24.5],
                                              [-65.5, -25]]])

        self.l8SR_col = ee.ImageCollection("LANDSAT/LC8_SR")

    def test_getRegion(self):
        expected = [[0.0,0.0], [10.0, 0.0], [10.0, 10.0], [0.0, 10.0], [0.0, 0.0]]
        pol = ee.Geometry.Polygon(expected)
        feat = ee.Feature(pol)

        region_geom = tools.getRegion(pol)
        region_feat = tools.getRegion(feat)

        self.assertEqual(region_geom, [expected])
        self.assertEqual(region_feat, [expected])

    def test_mask2(self):
        masked_img = self.l8SR.updateMask(self.l8SR.select(["cfmask"]).eq(0))

        mask2zero = tools.mask2zero(masked_img)
        val_no_change = tools.get_value(mask2zero, self.p_l8SR_no_cloud, 30)["B1"]
        val_change = tools.get_value(mask2zero, self.p_l8SR_cloud, 30)["B1"]

        self.assertEqual(val_no_change, 517)
        self.assertEqual(val_change, 0)

        mask2num = tools.mask2number(10)(masked_img)
        val_no_change_num = tools.get_value(mask2num, self.p_l8SR_no_cloud, 30)["B1"]
        val_change_num = tools.get_value(mask2num, self.p_l8SR_cloud, 30)["B1"]

        self.assertEqual(val_no_change_num, 517)
        self.assertEqual(val_change_num, 10)

    def test_addConstantBands(self):
        newimg = tools.addConstantBands(0, "a", "b", "c")(self.l8SR)
        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30)

        self.assertEqual(vals["B1"], 517)
        self.assertEqual(vals["a"], 0)
        self.assertEqual(vals["b"], 0)
        self.assertEqual(vals["c"], 0)

        newimg2 = tools.addConstantBands(a=0, b=1, c=2)(self.l8SR)
        vals2 = tools.get_value(newimg2, self.p_l8SR_no_cloud, 30)

        self.assertEqual(vals2["B1"], 517)
        self.assertEqual(vals2["a"], 0)
        self.assertEqual(vals2["b"], 1)
        self.assertEqual(vals2["c"], 2)

        newimg3 = tools.addConstantBands(0, "a", "b", "c", d=1, e=2)(self.l8SR)
        vals3 = tools.get_value(newimg3, self.p_l8SR_no_cloud, 30)

        self.assertEqual(vals3["B1"], 517)
        self.assertEqual(vals3["a"], 0)
        self.assertEqual(vals3["b"], 0)
        self.assertEqual(vals3["c"], 0)
        self.assertEqual(vals3["d"], 1)
        self.assertEqual(vals3["e"], 2)

    def test_replace(self):
        testband = ee.Image.constant(10).select([0],["anyname"])
        newimg = tools.replace(self.l8SR, "B1", testband)
        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30)
        self.assertEqual(vals["B1"], 10)

    def test_sumBands(self):
        newimg = tools.sumBands("added_bands", ("B1", "B2", "B3"))(self.l8SR)

        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30)
        suma = int(vals["B1"]) + int(vals["B2"]) + int(vals["B3"])

        self.assertEqual(vals["added_bands"], suma)

    def test_replace_many(self):
        pass

    def test_rename_bands(self):
        i = tools.rename_bands({"B1": "BLUE", "B2": "GREEN"})(self.l8SR)

        vals = tools.get_value(i, self.p_l8SR_no_cloud, 30)
        bands = i.bandNames().getInfo()

        self.assertEqual(bands, ["BLUE", "GREEN", "B3", "B4", "B5", "B6", "B7",
                                 "cfmask", "cfmask_conf"])
        self.assertEqual(vals["BLUE"], 517)

    def test_pass_prop(self):
        empty1 = ee.Image.constant(0)
        empty2 = ee.Image.constant(0).set("satellite", "None")

        pass1 = tools.pass_prop(self.l8SR, empty1, ["satellite", "WRS_PATH"])
        pass2 = tools.pass_prop(self.l8SR, empty2, ["satellite", "WRS_PATH"])

        self.assertEqual(pass1.get("satellite").getInfo(), "LANDSAT_8")
        self.assertEqual(pass2.get("satellite").getInfo(), "LANDSAT_8")
        self.assertEqual(pass1.get("WRS_PATH").getInfo(), 231)
        self.assertEqual(pass2.get("WRS_PATH").getInfo(), 231)

    def test_list_intersection(self):
        intersection = tools.list_intersection(self.list1, self.list2).getInfo()
        self.assertEqual(intersection, [4, 5])

    def test_list_diff(self):
        diff = tools.list_diff(self.list1, self.list2).getInfo()
        self.assertEqual(diff, [1, 2, 3, 6, 7])

    def test_parametrize(self):
        newimg = tools.parametrize((0, 10000), (0, 1), ["B1", "B2"])(self.l8SR)

        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30)
        self.assertEqual(vals["B1"], 517.0/10000)
        self.assertEqual(vals["B2"], 580.0/10000)
        self.assertEqual(vals["B3"], 824)

    def test_cloud_mask(self):
        from geetools import cloud_mask
        # LANDSAT 4 TOA
        masked_l4toa = cloud_mask.fmask("fmask")(self.l4toa)
        p_l4toa = ee.Geometry.Point([-72.2736, -44.6062])
        vals_l4toa = tools.get_value(masked_l4toa, p_l4toa, 30)
        self.assertEqual(vals_l4toa["B1"], None)

        # LANDSAT 5 USGS
        p_l5usgs = ee.Geometry.Point([-65.4991, -24.534])
        masked_l5usgs = cloud_mask.fmask("cfmask")(self.l5SR)
        vals_l5usgs = tools.get_value(masked_l5usgs, p_l5usgs, 30)
        self.assertEqual(vals_l5usgs["B1"], None)

        # LANDSAT 5 LEDAPS
        p_l5led = ee.Geometry.Point([-70.3455, -43.8306])
        masked_l5led = cloud_mask.ledaps(self.l5led)
        vals_l5led = tools.get_value(masked_l5led, p_l5led, 30)
        self.assertEqual(vals_l5led["B1"], None)

        # LANDSAT 5 TOA
        masked_l5toa = cloud_mask.fmask("fmask")(self.l5toa)
        p_l5toa = ee.Geometry.Point([-70.9003, -39.7421])
        vals_l5toa = tools.get_value(masked_l5toa, p_l5toa, 30)
        self.assertEqual(vals_l5toa["B1"], None)

        # LANDSAT 7 USGS
        p_l7usgs = ee.Geometry.Point([-64.8533, -26.1052])
        masked_l7usgs = cloud_mask.fmask("cfmask")(self.l7SR)
        vals_l7usgs = tools.get_value(masked_l7usgs, p_l7usgs, 30)
        self.assertEqual(vals_l7usgs["B1"], None)

        # LANDSAT 7 LEDAPS
        p_l7led = ee.Geometry.Point([-64.9141, -25.2264])
        masked_l7led = cloud_mask.ledaps(self.l7led)
        vals_l7led = tools.get_value(masked_l7led, p_l7led, 30)
        self.assertEqual(vals_l7led["B1"], None)

        # LANDSAT 7 TOA
        masked_l7toa = cloud_mask.fmask("fmask")(self.l7toa)
        p_l7toa = ee.Geometry.Point([-64.8495, -25.2354])
        vals_l7toa = tools.get_value(masked_l7toa, p_l7toa, 30)
        self.assertEqual(vals_l7toa["B1"], None)

        # LANDSAT 8 USGS
        p_l8usgs = ee.Geometry.Point([-65.5568, -24.3327])
        masked_l8usgs = cloud_mask.fmask("cfmask")(self.l8SR)
        vals_l8usgs = tools.get_value(masked_l8usgs, p_l8usgs, 30)
        self.assertEqual(vals_l8usgs["B1"], None)

        # LANDSAT 8 TOA
        masked_l8toa = cloud_mask.fmask("fmask")(self.l8toa)
        p_l8toa = ee.Geometry.Point([-71.7833, -43.6634])
        vals_l8toa = tools.get_value(masked_l8toa, p_l8toa, 30)
        self.assertEqual(vals_l8toa["B1"], None)

        # LANDSAT SENTINEL 2
        masked_s2 = cloud_mask.sentinel2(self.sentinel2)
        p_s2 = ee.Geometry.Point([-72.2104, -42.7592])
        vals_s2 = tools.get_value(masked_s2, p_s2, 10)
        self.assertEqual(vals_s2["B1"], None)

        # LANDSAT MODIS AQ
        masked_ma = cloud_mask.modis(self.modis_aqua)
        p_ma = ee.Geometry.Point([-69.071, -43.755])
        vals_ma = tools.get_value(masked_ma, p_ma, 500)
        self.assertEqual(vals_ma["sur_refl_b01"], None)

        # LANDSAT MODIS TE
        masked_te = cloud_mask.modis(self.modis_terra)
        p_te = ee.Geometry.Point([-71.993, -43.692])
        vals_te = tools.get_value(masked_te, p_te, 500)
        self.assertEqual(vals_te["sur_refl_b01"], None)

    def test_expressions(self):
        from geetools import expressions
        generator = expressions.ExpGen()
        exp_max = generator.max("b('B1')", "b('B2')")
        exp_min = generator.min("b('B1')", "b('B2')")

        img_max = self.l8SR.expression(exp_max).select([0], ["max"])
        img_min = self.l8SR.expression(exp_min).select([0], ["min"])

        vals_max = tools.get_value(img_max, self.p_l8SR_no_cloud, 30)
        vals_min = tools.get_value(img_min, self.p_l8SR_no_cloud, 30)

        self.assertEqual(vals_max["max"], 580)
        self.assertEqual(vals_min["min"], 517)



# coding=utf-8
import unittest
import ee
from .. import tools
ee.Initialize()

class TestTools(unittest.TestCase):

    def setUp(self):
        self.l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")

        self.p_l8SR_cloud = ee.Geometry.Point([-65.8109, -25.0185])
        self.p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

        self.list1 = ee.List([1,2,3,4,5])
        self.list2 = ee.List([4,5,6,7])

    def test_getRegion(self):
        expected = [[0.0,0.0], [10.0, 0.0], [10.0, 10.0], [0.0, 10.0], [0.0, 0.0]]
        pol = ee.Geometry.Polygon(expected)
        feat = ee.Feature(pol)

        region_geom = tools.getRegion(pol)
        region_feat = tools.getRegion(feat)

        self.assertEqual(region_geom, [expected])
        self.assertEqual(region_feat, [expected])

    def test_mask2(self):
        masked_img = self.l8SR.updateMask(self.l8SR.select(["cfmask"]).eq(0))

        mask2zero = tools.mask2zero(masked_img)
        val_no_change = tools.get_value(mask2zero, self.p_l8SR_no_cloud, 30, 'client')["B1"]
        val_change = tools.get_value(mask2zero, self.p_l8SR_cloud, 30, 'client')["B1"]

        self.assertEqual(val_no_change, 517)
        self.assertEqual(val_change, 0)

        mask2num = tools.mask2number(10)(masked_img)
        val_no_change_num = tools.get_value(mask2num, self.p_l8SR_no_cloud, 30, 'client')["B1"]
        val_change_num = tools.get_value(mask2num, self.p_l8SR_cloud, 30, 'client')["B1"]

        self.assertEqual(val_no_change_num, 517)
        self.assertEqual(val_change_num, 10)

    def test_addConstantBands(self):
        newimg = tools.addConstantBands(0, "a", "b", "c")(self.l8SR)
        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30, 'client')

        self.assertEqual(vals["B1"], 517)
        self.assertEqual(vals["a"], 0)
        self.assertEqual(vals["b"], 0)
        self.assertEqual(vals["c"], 0)

        newimg2 = tools.addConstantBands(a=0, b=1, c=2)(self.l8SR)
        vals2 = tools.get_value(newimg2, self.p_l8SR_no_cloud, 30, 'client')

        self.assertEqual(vals2["B1"], 517)
        self.assertEqual(vals2["a"], 0)
        self.assertEqual(vals2["b"], 1)
        self.assertEqual(vals2["c"], 2)

        newimg3 = tools.addConstantBands(0, "a", "b", "c", d=1, e=2)(self.l8SR)
        vals3 = tools.get_value(newimg3, self.p_l8SR_no_cloud, 30, 'client')

        self.assertEqual(vals3["B1"], 517)
        self.assertEqual(vals3["a"], 0)
        self.assertEqual(vals3["b"], 0)
        self.assertEqual(vals3["c"], 0)
        self.assertEqual(vals3["d"], 1)
        self.assertEqual(vals3["e"], 2)

    def test_replace(self):
        testband = ee.Image.constant(10).select([0],["anyname"])
        newimg = tools.replace(self.l8SR, "B1", testband)
        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30, 'client')
        self.assertEqual(vals["B1"], 10)

    def test_sumBands(self):
        newimg = tools.sumBands("added_bands", ("B1", "B2", "B3"))(self.l8SR)

        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30, 'client')
        suma = int(vals["B1"]) + int(vals["B2"]) + int(vals["B3"])

        self.assertEqual(vals["added_bands"], suma)

    def test_replace_many(self):
        pass

    def test_rename_bands(self):
        i = tools.rename_bands({"B1": "BLUE", "B2": "GREEN"})(self.l8SR)

        vals = tools.get_value(i, self.p_l8SR_no_cloud, 30, 'client')
        bands = i.bandNames().getInfo()

        self.assertEqual(bands, ["BLUE", "GREEN", "B3", "B4", "B5", "B6", "B7",
                                 "cfmask", "cfmask_conf"])
        self.assertEqual(vals["BLUE"], 517)

    def test_pass_prop(self):
        empty1 = ee.Image.constant(0)
        empty2 = ee.Image.constant(0).set("satellite", "None")

        pass1 = tools.pass_prop(self.l8SR, empty1, ["satellite", "WRS_PATH"])
        pass2 = tools.pass_prop(self.l8SR, empty2, ["satellite", "WRS_PATH"])

        self.assertEqual(pass1.get("satellite").getInfo(), "LANDSAT_8")
        self.assertEqual(pass2.get("satellite").getInfo(), "LANDSAT_8")
        self.assertEqual(pass1.get("WRS_PATH").getInfo(), 231)
        self.assertEqual(pass2.get("WRS_PATH").getInfo(), 231)

    def test_list_intersection(self):
        intersection = tools.list_intersection(self.list1, self.list2).getInfo()
        self.assertEqual(intersection, [4, 5])

    def test_list_diff(self):
        diff = tools.list_diff(self.list1, self.list2).getInfo()
        self.assertEqual(diff, [1, 2, 3, 6, 7])

    def test_parametrize(self):
        newimg = tools.parametrize((0, 10000), (0, 1), ["B1", "B2"])(self.l8SR)

        vals = tools.get_value(newimg, self.p_l8SR_no_cloud, 30, 'client')
        self.assertEqual(vals["B1"], 517.0/10000)
        self.assertEqual(vals["B2"], 580.0/10000)
        self.assertEqual(vals["B3"], 824)


# coding=utf-8
''' Creation of strip of images '''

from __future__ import print_function
from . import tools
from PIL import Image as ImPIL
from PIL import ImageDraw, ImageFont
import os.path
import ee
import logging

def listEE2list(listEE, type='Image'):
    relation = {'Image': ee.Image,
                'Number': ee.Number,
                'String': ee.String,
                'Feature': ee.Feature}
    size = listEE.size().getInfo()
    newlist = []
    for el in range(size):
        newlist.append(relation[type](listEE.get(el)))

    return newlist

class ImageStrip(object):
    ''' Create an image strip '''
    def __init__(self, name, ext="png", description="", **kwargs):
        '''
        :param name: File name
        :type name: str

        :Opcionals:

        :param ext: Extention. Default: png
        :type ext: str
        :param body_size: Body size. Defaults to 18
        :type body_size: int
        :param title_size: Title's font size. Defaults to 30
        :type title_size: int
        :param font: Font for all texts. defaults to "DejaVuSerif.ttf"
        :type font: str
        :param background_color: Background color. defaults to "white"
        :type background_color: str
        :param title_color: Title's font color. defaults to "black"
        :type title_color: str
        :param body_color: Body's text color. defaults to "red"
        :type body_color:  str
        :param general_width: defaults to 0
        :type general_width: int
        :param general_height: defaults to 0
        :type general_height: int
        :param y_space: Space between lines or rows. defaults to 10
        :type y_space: int
        :param x_space: Space between columns. defaults to 15
        :type x_space: int
        :param description: Description that will go beneath the title
        :type description: str
        '''
        self.name = name
        self.ext = ext
        self.description = description

        self.body_size = kwargs.get("body_size", 18)
        self.title_size = kwargs.get("title_size", 30)
        self.font = kwargs.get("font", "DejaVuSerif.ttf")
        self.background_color = kwargs.get("background_color", "white")
        self.title_color = kwargs.get("title_color", "black")
        self.body_color = kwargs.get("body_color", "red")
        self.general_width = kwargs.get("general_width", 0)
        self.general_height = kwargs.get("general_height", 0)
        self.y_space = kwargs.get("y_space", 10)
        self.x_space = kwargs.get("x_space", 15)

    @staticmethod
    def unpack(doublelist):
        return [y for x in doublelist for y in x]

    def create(self, imlist, namelist, desclist=None):
        """ Main method to create the actual strip
        
        :param imlist: PIL images, ej: [[img1, img2],[img3, img4]]
        :type imlist: list of lists
        :param namelist: Names for the images. Must match imlist size
        :type namelist: list of lists
        :param desclist: Descriptions for the images. Must match imlist size
        :type desclist: list of lists
        :return:
        :rtype:
        """

        # SANITY CHECK
        err1 = "dimension of imlist must match dimension of namelist"
        err2 = "dimension of nested lists must be equal"

        if len(imlist) != len(namelist):
            raise ValueError(err1)
        elif [len(l) for l in imlist] != [len(l) for l in namelist]:
            raise ValueError(err2)

        if desclist is not None:
            if len(desclist) != len(imlist):
                raise ValueError(err1)
            elif [len(l) for l in imlist] != [len(l) for l in desclist]:
                raise ValueError(err2)

        def line_height(text, font):
            """ Calculate height for a multiline text
            
            :param text: multiline text
            :type text: str
            :param font: utilized font
            :type font: ImageFont
            :return: text height
            :rtype: float
            """
            lista = text.split("\n")
            alt = 0
            for linea in lista:
                alt += font.getsize(linea)[1]
            return alt + self.y_space

        # FUENTES
        font = ImageFont.truetype(self.font, self.body_size)
        fontit = ImageFont.truetype(self.font, self.title_size)
        font_desc = ImageFont.truetype(self.font, self.title_size - 4)

        # TITULO
        title = self.name
        title_height = line_height(title, fontit)
        
        # DIMENSIONES DESCRIPCION
        description = self.description.replace("_", "\n")
        desc_height = line_height(description, font_desc)
        desc_width = font_desc.getsize(description)[0]

        # ALTURA NOMBRES
        if desclist:
            all_names = [n.replace("_", "\n") for n in self.unpack(desclist)]
        else:
            all_names = [n.replace("_", "\n") for n in self.unpack(namelist)]

        all_height = [line_height(t, font) for t in all_names]
        name_height = max(*all_height) #+ 2
        # print alturatodos, alturanombres

        # CALCULO EL ANCHO DE LA PLANTILLA
        imgs_width = [[ii.size[0] + self.x_space for ii in i] for i in imlist]
        imgs_width_sum = [sum(i) for i in imgs_width]

        # CALCULO EL MAXIMO ANCHO DE LA LISTA DE ANCHOS
        max_width = max(imgs_width_sum)
        strip_width = int(max_width)

        # CALCULO EL ALTO MAX

        # POR AHORA LO CALCULA CON EL ALTO DE LA 1ER COLUMNA
        # PERO SE PODRIA CALCULAR CUAL ES LA IMG MAS ALTA Y CUAL
        # ES EL TITULO MAS ALTO, Y COMBINAR AMBOS...

        # Tener en cuenta que listaimgs es una lista de listas [[i1,i2], [i3,i4]]

        # print "calculando general_height de la plantilla..."

        imgs_height = [[ii.size[1] for ii in i] for i in imlist]
        max_imgs_height = [max(*i) + name_height + self.y_space for i in imgs_height]
        max_height = sum(max_imgs_height)

        strip_height = int(sum((title_height, desc_height, max_height,
                                    self.y_space * 3)))

        strip = ImPIL.new("RGB", (strip_width, strip_height),
                              self.background_color)

        draw = ImageDraw.Draw(strip)

        # DIBUJA LOS ELEMENTOS DENTRO DE LA PLANTILLA

        # COORD INICIALES
        x = 0
        y = 0

        # DIBUJA EL TITULO
        draw.text((x, y), title, font=fontit, fill=self.title_color)

        # DIBUJA LA DESCRIPCION
        y += title_height + self.y_space # aumento y
        # print "y de la descripcion:", y
        draw.text((x, y), description, font=font_desc, fill=self.title_color)

        # DIBUJA LAS FILAS

        # logging.debug(("altura de la descripcion (calculada):", desc_height))
        # logging.debug(("altura de la desc", font_desc.getsize(description)[1]))
        y += desc_height + self.y_space # aumento y

        # DIBUJA LAS FILAS Y COLUMNAS

        if desclist:
            namelist = desclist

        for i, n, alto in zip(imlist, namelist, max_imgs_height):
            # RESETEO LA POSICION HORIZONTAL
            xn = x # hago esto porque en cada iteracion aumento solo xn y desp vuelvo a x
            for image, name in zip(i, n):
                # LA COLUMNA ES: (imagen, name, anchocolumna)
                # print image, name

                ancho_i, alto_i = image.size

                # DIBUJO imagen
                strip.paste(image, (xn, y))

                # aumento y para pegar el texto
                _y = y + alto_i + self.y_space

                # DIBUJO name
                draw.text((xn, _y),
                          name.replace("_", "\n"),
                          font=font,
                          fill=self.body_color)

                # AUMENTO y
                xn += ancho_i + self.x_space

            # AUMENTO x
            y += alto

        strip.save(self.name + "." + self.ext)
        strip.show()

        return strip

    def from_list(self, imlist, namelist, viz_bands, min, max, region,
                  folder, check=True, desclist=None, **kwargs):
        """ Download every image and create the strip

        :param imlist: Satellite Images (not PIL!!!!!!)
        :type imlist: list of ee.Image
        :param namelist: Names for the images. Must match imlist
        :type namelist: list of str
        :param desclist: list of descriptions for every image. Optional
        :type desclist: list of str
        :param viz_bands: Visualization bands. ej: ["NIR","SWIR","RED"]
        :type viz_bands: list of str
        :param min: min value for visualization
        :type min: int
        :param max: max value for visualization
        :type max: int
        :param region: coordinate list. Optional
        :type region: list
        :param folder: folder to downlaod files. Do not use '/' at the end
        :type folder: str
        :param check: Check if file exists, and if it does, omits the downlaod
        :type check: bool
        :param draw_polygons: Polygons to draw over the image. Must be a list of list of
            coordinates. Optional
        :type draw_polygons: list of list
        :param draw_lines: Lines to draw over the image
        :type draw_lines: list of list
        :param draw_points: Points to draw over the image
        :type draw_points: list of list
        :param general_width: Images width
        :type general_width: int

        :return: A file with the name passed to StripImage() in the folder
            passed to the method. Opens the generated file
        """

        if isinstance(imlist, ee.List):
            imlist = listEE2list(imlist, 'Image')

        region = tools.execli(
            ee.Geometry.Polygon(region).bounds().getInfo)()["coordinates"]

        # TODO: modificar este metodo para que se pueda pasar listas de listas
        general_width = kwargs.get("general_width", 500)

        draw_polygons = kwargs.get("draw_polygons", None)
        draw_lines = kwargs.get("draw_lines", None)
        draw_points = kwargs.get("draw_points", None)

        # desclist = kwargs.get("desclist", None)

        list_of_lists = []

        # logging.debug(("verificar archivo?", check))

        # for i in range(0, len(imlist)):
        for img_list, nom_list in zip(imlist, namelist):
            pil_img_list = []
            for image, name in zip(img_list, nom_list):
                # name = namelist[i]
                # exist = os.path.exists(folder+"/"+name)
                # path = "{0}/{1}.{2}".format(folder, name, self.ext)

                # CHECK CARPETA
                abscarp = os.path.abspath(folder)
                if not os.path.exists(abscarp):
                    os.mkdir(abscarp)

                path = "{}/{}".format(folder, name)
                fullpath = os.path.abspath(path)+"."+self.ext
                exist = os.path.isfile(fullpath)

                logging.debug(("existe {}?".format(fullpath), exist))

                if check and exist:
                    im = ImPIL.open(fullpath)
                else:
                    img = ee.Image(image)
                    urlviz = img.visualize(bands=viz_bands, min=min, max=max,
                                           forceRgbOutput=True)

                    url = urlviz.getThumbURL({"region": region,
                                              "format": self.ext,
                                              "dimensions": general_width})

                    # archivo = funciones.downloadFile3(url, folder+"/"+name, self.ext)
                    file = tools.downloadFile(url, path, self.ext)

                    im = ImPIL.open(file.name)
                    # listaimPIL.append(im)

                dr = ImageDraw.Draw(im)

                general_width, general_height = im.size
                # print "general_width:", general_width, "general_height:", general_height

                # nivel de 'anidado' (nested)
                def nested_level(l):
                    n = 0
                    while type(l[0]) is list:
                        n += 1
                        l = l[0]
                    return n

                region = region[0] if nested_level(region) == 2 else region
                region = region[:-1] if len(region) == 5 else region

                p0 = region[0]
                p1 = region[1]
                p3 = region[3]
                distX = abs(p0[0]-p1[0])
                distY = abs(p0[1]-p3[1])

                # FACTORES DE ESCALADO
                width_ratio = float(general_width)/float(distX)
                height_ratio = float(general_height)/float(distY)

                if draw_polygons:
                    for pol in draw_polygons:
                        pol = [tuple(p) for p in pol]

                        newpol = [(abs(x-p0[0])*width_ratio, abs(y-p0[1])*height_ratio) for x, y in pol]

                        # logging.debug("nuevas coords {}".format(newpol))

                        #print "\n\n", pol, "\n\n"
                        dr.polygon(newpol, outline="red")

                pil_img_list.append(im)
            list_of_lists.append(pil_img_list)

        if desclist:
            return self.create(list_of_lists, namelist, desclist)
        else:
            return self.create(list_of_lists, namelist)

    def from_collection(self, collections, viz_param=None, region=None,
                        name=None, folder=None, properties=None,
                        drawRegion=False, zoom=0, description="", **kwargs):
        ''' Create an image strip from a given ee.ImageCollection or ee.List

        :param collections: contains the images (ee.Image)
        :type collections: list of ee.ImageCollection or ee.List
        :param viz_param: visualization parameters. If None will use data of
            the first 3 bands.
        :type viz_param: dict
        :param region: region from where to create the image strip. If None,
            the region of the first image is used
        :type region: list of list
        :param name: name for the folder. If None the name of the
            collection will be used
        :type name: str
        :param folder: folder to save the file. If None, the name of the
            collection will be used
        :type folder: str
        :param properties: properties to add as a description of every image
        :type properties: list
        :param drawRegion: draw a polygon in the given region
        :type drawRegion: bool
        :param zoom: doesn't match EE zoom, this strats from 0 (no zoom).
        :type zoom: int
        :param description: Paragraph that serves as a decription of the image
            strip
        :type description: str
        :param kwargs: other parameters passed to the creation of the
            ImageStrip object.
        :type kwargs: dict
        :return: A file with the name passed to StripImage() in the folder
            passed to the method. Opens the generated file
        '''
        # SANITY CHECK
        coltypes_imcol = [isinstance(col, ee.ImageCollection) for col in collections]
        coltypes_imlist = [isinstance(col, ee.List) for col in collections]
        if not min(coltypes_imcol) and not min(coltypes_imlist):
            raise ValueError("'collection' parameter must be a list of " \
                "ee.ImageCollection or a list of ee.List containing ee.Image")

        if folder is None:
            folder = name

        # TOMA DATOS DE LA PRIMER IMAGEN POR SI NO SE ESPECIFICA LA REGION O LOS
        # PARAMETROS PARA LA VISUALIZACION
        first = ee.Image(collections[0].first())
        first_info = tools.execli(first.getInfo)()

        # OBTENGO LA REGION, DE ESTE MODO SE PUEDE ESPECIFICAR UN AREA MAS CHICA
        if not region:
            region = first.geometry().bounds().getInfo()["coordinates"]

        original_region = region

        maxError = kwargs.get("maxError", 10)
        if zoom > 0:
            pol = ee.Geometry.Polygon(region)
            center = pol.centroid()
            b = center.buffer(zoom*100, maxError).bounds()
            region = tools.execli(b.getInfo)()["coordinates"]

        if viz_param:
            bands = viz_param.get("bands")
            minV = viz_param.get("min")
            maxV = viz_param.get("max")
        else:
            bandtype = first_info["bands"][0]["data_type"]["precision"]
            bandmax = first_info["bands"][0]["data_type"]["max"]
            b1 = first_info["bands"][0]["id"]
            b2 = first_info["bands"][1]["id"]
            b3 = first_info["bands"][2]["id"]
            bands = [b1, b2, b3]
            minV = 0
            if bandtype == "int" and bandmax > 255:
                maxV = 5000
            elif bandtype == "float":
                maxV = 0.5
            elif bandtype == "int" and bandmax == 255:
                maxV = 125
            else:
                maxV = 0.5

        zoom_str = "z"+str(zoom)
        region_str = "with_region" if drawRegion else "without_region"

        name += "_" + "_".join((zoom_str, region_str))

        pl = ImageStrip(name, description=description)

        nested_imgs = []
        nested_names = []
        nested_desc = []

        for col in collections:
            # TRANSFORMO LA COLECCION EN UNA LISTA
            lista = col.toList(5000)
            listaPy = []
            names = []
            descriptions = [] # description para cada imagen
            size = tools.execli(lista.size().getInfo)()
            # colID = funciones.execli(col.getInfo)()["id"].split("/")[-1]

            for i in range(0, size):
                img = ee.Image(lista.get(i))
                listaPy.append(img)
                date = tools.execli(img.date().format().getInfo)()
                imID = tools.execli(img.id().getInfo)()
                propname = ""
                if properties:
                    for prop in properties:
                        value = tools.execli(img.get(prop).getInfo)()
                        value = str(value) if value else "no {} property".format(value)
                        propname += "{}: {}_".format(prop, value)

                # desc = "{fecha}".format(id=imID, fecha=date)
                desc = "ID: {imID}_date: {date}_{propname}".format(**locals())
                # nn = "_".join((colID, str(date), zoom_str, buff_str))
                nn = "_".join((imID, str(date), zoom_str))

                names.append(nn)
                descriptions.append(desc)

            # logging.debug(("descripciones", descripciones))
            # logging.debug(("nombres", nombres))

            nested_imgs.append(listaPy)
            nested_names.append(names)
            nested_desc.append(descriptions)

        arguments = dict(
            imlist=nested_imgs,
            namelist=nested_names,
            desclist=nested_desc,
            viz_bands=bands,
            min=minV,
            max=maxV,
            region=region,
            folder=folder,
            check=True)

        if drawRegion:
            arguments["draw_polygons"] = (original_region)

        # print arguments["draw_polygons"]

        i = pl.from_list(**arguments)

        return i


#!/usr/bin/env python

import os
from setuptools import setup, find_packages
from geetools import __version__


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# the setup
setup(
    name='geetools',
    version=__version__,
    description='Set of tools to use in Google Earth Engine Python API',
    long_description=read('README.rst'),
    url='',
    author='Rodrigo E. Principe',
    author_email='fitoprincipe82@gmail.com',
    license='MIT',
    keywords='google earth engine raster image processing gis satelite',
    packages=find_packages(exclude=('docs', 'js')),
    include_package_data=True,
    install_requires=['requests', 'Pillow', 'simpleeval', 'pyshp',
                      'folium', 'ipyleaflet', 'traitlets', 'pygal'],
    extras_require={
    'dev': [],
    'docs': [],
    'testing': [],
    },
    classifiers=['Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7'],
    )



# coding=utf-8
import unittest
import ee
from .. import tools
ee.Initialize()

class TestExp(unittest.TestCase):

    def setUp(self):
        self.l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")
        self.p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

    def test_expressions(self):
        from geetools import expressions
        generator = expressions.ExpGen()
        exp_max = generator.max("b('B1')", "b('B2')")
        exp_min = generator.min("b('B1')", "b('B2')")

        img_max = self.l8SR.expression(exp_max).select([0], ["max"])
        img_min = self.l8SR.expression(exp_min).select([0], ["min"])

        vals_max = tools.get_value(img_max, self.p_l8SR_no_cloud, 30, 'client')
        vals_min = tools.get_value(img_min, self.p_l8SR_no_cloud, 30, 'client')

        self.assertEqual(vals_max["max"], 580)
        self.assertEqual(vals_min["min"], 517)